__author__ = "Simon Nilsson"

import os.path
import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
import atexit
import subprocess
import sys
import threading
import tkinter.ttk as ttk
import urllib.request
import webbrowser
from tkinter.filedialog import askdirectory, askopenfilename
from tkinter.messagebox import askyesno

import PIL.Image
from PIL import ImageTk

from simba.bounding_box_tools.boundary_menus import BoundaryMenus
from simba.cue_light_tools.cue_light_menues import CueLightAnalyzerMenu
from simba.labelling.labelling_advanced_interface import \
    select_labelling_video_advanced
from simba.labelling.labelling_interface import select_labelling_video
from simba.labelling.targeted_annotations_clips import \
    select_labelling_video_targeted_clips
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.model.grid_search_rf import GridSearchRandomForestClassifier
from simba.model.inference_batch import InferenceBatch
from simba.model.inference_validation import InferenceValidation
from simba.model.train_rf import TrainRandomForestClassifier
from simba.outlier_tools.outlier_corrector_location import \
    OutlierCorrecterLocation
from simba.outlier_tools.outlier_corrector_movement import \
    OutlierCorrecterMovement
from simba.outlier_tools.skip_outlier_correction import \
    OutlierCorrectionSkipper
from simba.plotting.interactive_probability_grapher import \
    InteractiveProbabilityGrapher
from simba.roi_tools.ROI_define import *
from simba.roi_tools.ROI_menus import *
from simba.roi_tools.ROI_reset import *
from simba.third_party_label_appenders.BENTO_appender import BentoAppender
from simba.third_party_label_appenders.BORIS_appender import BorisAppender
from simba.third_party_label_appenders.deepethogram_importer import \
    DeepEthogramImporter
from simba.third_party_label_appenders.ethovision_import import \
    ImportEthovision
from simba.third_party_label_appenders.observer_importer import \
    NoldusObserverImporter
from simba.third_party_label_appenders.solomon_importer import SolomonImporter
from simba.ui.create_project_ui import ProjectCreatorPopUp
from simba.ui.machine_model_settings_ui import MachineModelSettingsPopUp
from simba.ui.pop_ups.about_simba_pop_up import AboutSimBAPopUp
from simba.ui.pop_ups.animal_directing_other_animals_pop_up import \
    AnimalDirectingAnimalPopUp
from simba.ui.pop_ups.append_roi_features_animals_pop_up import \
    AppendROIFeaturesByAnimalPopUp
from simba.ui.pop_ups.append_roi_features_bodypart_pop_up import \
    AppendROIFeaturesByBodyPartPopUp
from simba.ui.pop_ups.archive_files_pop_up import ArchiveProcessedFilesPopUp
from simba.ui.pop_ups.batch_preprocess_pop_up import BatchPreProcessPopUp
from simba.ui.pop_ups.boolean_conditional_slicer_pup_up import \
    BooleanConditionalSlicerPopUp
from simba.ui.pop_ups.clf_add_remove_print_pop_up import (
    AddClfPopUp, PrintModelInfoPopUp, RemoveAClassifierPopUp)
from simba.ui.pop_ups.clf_by_roi_pop_up import ClfByROIPopUp
from simba.ui.pop_ups.clf_by_timebins_pop_up import TimeBinsClfPopUp
from simba.ui.pop_ups.clf_descriptive_statistics_pop_up import \
    ClfDescriptiveStatsPopUp
from simba.ui.pop_ups.clf_plot_pop_up import SklearnVisualizationPopUp
from simba.ui.pop_ups.clf_probability_plot_pop_up import \
    VisualizeClassificationProbabilityPopUp
from simba.ui.pop_ups.clf_validation_plot_pop_up import \
    ClassifierValidationPopUp
from simba.ui.pop_ups.csv_2_parquet_pop_up import (Csv2ParquetPopUp,
                                                   Parquet2CsvPopUp)
from simba.ui.pop_ups.data_plot_pop_up import DataPlotterPopUp
from simba.ui.pop_ups.directing_animal_to_bodypart_plot_pop_up import \
    DirectingAnimalToBodyPartVisualizerPopUp
from simba.ui.pop_ups.directing_other_animals_plot_pop_up import \
    DirectingOtherAnimalsVisualizerPopUp
from simba.ui.pop_ups.direction_animal_to_bodypart_settings_pop_up import \
    DirectionAnimalToBodyPartSettingsPopUp
from simba.ui.pop_ups.distance_plot_pop_up import DistancePlotterPopUp
from simba.ui.pop_ups.fsttc_pop_up import FSTTCPopUp
from simba.ui.pop_ups.gantt_pop_up import GanttPlotPopUp
from simba.ui.pop_ups.heatmap_clf_pop_up import HeatmapClfPopUp
from simba.ui.pop_ups.heatmap_location_pop_up import HeatmapLocationPopup
from simba.ui.pop_ups.kleinberg_pop_up import KleinbergPopUp
from simba.ui.pop_ups.make_path_plot_pop_up import MakePathPlotPopUp
from simba.ui.pop_ups.movement_analysis_pop_up import MovementAnalysisPopUp
from simba.ui.pop_ups.movement_analysis_time_bins_pop_up import \
    MovementAnalysisTimeBinsPopUp
from simba.ui.pop_ups.mutual_exclusivity_pop_up import MutualExclusivityPupUp
from simba.ui.pop_ups.outlier_settings_pop_up import OutlierSettingsPopUp
from simba.ui.pop_ups.path_plot_pop_up import PathPlotPopUp
from simba.ui.pop_ups.pose_bp_drop_pop_up import DropTrackingDataPopUp
from simba.ui.pop_ups.pose_reorganizer_pop_up import PoseReorganizerPopUp
from simba.ui.pop_ups.pup_retrieval_pop_up import PupRetrievalPopUp
from simba.ui.pop_ups.quick_path_plot_pop_up import QuickLineplotPopup
from simba.ui.pop_ups.remove_roi_features_pop_up import RemoveROIFeaturesPopUp
from simba.ui.pop_ups.roi_analysis_pop_up import ROIAnalysisPopUp
from simba.ui.pop_ups.roi_analysis_time_bins_pop_up import \
    ROIAnalysisTimeBinsPopUp
from simba.ui.pop_ups.roi_features_plot_pop_up import VisualizeROIFeaturesPopUp
from simba.ui.pop_ups.roi_size_standardizer_popup import \
    ROISizeStandardizerPopUp
from simba.ui.pop_ups.roi_tracking_plot_pop_up import VisualizeROITrackingPopUp
from simba.ui.pop_ups.set_machine_model_parameters_pop_up import \
    SetMachineModelParameters
from simba.ui.pop_ups.severity_analysis_pop_up import AnalyzeSeverityPopUp
from simba.ui.pop_ups.smoothing_interpolation_pop_up import (InterpolatePopUp,
                                                             SmoothingPopUp)
from simba.ui.pop_ups.subset_feature_extractor_pop_up import \
    FeatureSubsetExtractorPopUp
from simba.ui.pop_ups.third_party_annotator_appender_pop_up import \
    ThirdPartyAnnotatorAppenderPopUp
from simba.ui.pop_ups.validation_plot_pop_up import ValidationVideoPopUp
from simba.ui.pop_ups.video_processing_pop_up import (
    CalculatePixelsPerMMInVideoPopUp, ChangeFpsMultipleVideosPopUp,
    ChangeFpsSingleVideoPopUp, ChangeImageFormatPopUp, CLAHEPopUp,
    ClipVideoPopUp, ConcatenatingVideosPopUp, ConcatenatorPopUp,
    ConvertROIDefinitionsPopUp, ConvertVideoPopUp, CreateGIFPopUP,
    CropVideoPopUp, DownsampleVideoPopUp, ExtractAllFramesPopUp,
    ExtractAnnotationFramesPopUp, ExtractSEQFramesPopUp,
    ExtractSpecificFramesPopUp, GreyscaleSingleVideoPopUp,
    ImportFrameDirectoryPopUp, MergeFrames2VideoPopUp, MultiCropPopUp,
    MultiShortenPopUp, SuperImposeFrameCountPopUp, VideoRotatorPopUp,
    VideoTemporalJoinPopUp)
from simba.ui.pop_ups.visualize_pose_in_dir_pop_up import \
    VisualizePoseInFolderPopUp
from simba.ui.tkinter_functions import DropDownMenu, Entry_Box, FileSelect
from simba.ui.video_info_ui import VideoInfoTable
from simba.utils.checks import (check_ffmpeg_available,
                                check_file_exist_and_readable, check_int)
from simba.utils.custom_feature_extractor import CustomFeatureExtractor
from simba.utils.data import run_user_defined_feature_extraction_class
from simba.utils.enums import OS, Defaults, Formats, TagNames
from simba.utils.errors import InvalidInputError
from simba.utils.lookups import (get_bp_config_code_class_pairs, get_emojis,
                                 get_icons_paths)
from simba.utils.printing import stdout_success, stdout_warning
from simba.utils.read_write import get_video_meta_data
from simba.utils.warnings import FFMpegNotFoundWarning, PythonVersionWarning
from simba.video_processors.video_processing import (
    extract_frames_from_all_videos_in_directory, superimpose_frame_count,
    video_to_greyscale)

# from simba.unsupervised.ui import UnsupervisedGUI

sys.setrecursionlimit(10**6)
currentPlatform = platform.system()


class LoadProjectPopUp(object):
    def __init__(self):
        self.main_frm = Toplevel()
        self.main_frm.minsize(300, 200)
        self.main_frm.wm_title("Load SimBA project (project_config.ini file)")
        self.load_project_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="LOAD SIMBA PROJECT_CONFIG.INI",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.LOAD_PROJECT.value,
        )
        self.selected_file = FileSelect(
            self.load_project_frm,
            "Select file: ",
            title="Select project_config.ini file",
            file_types=[("SimBA Project .ini", "*.ini")],
        )
        load_project_btn = Button(
            self.load_project_frm,
            text="LOAD PROJECT",
            fg="blue",
            command=lambda: self.launch_project(),
        )

        self.load_project_frm.grid(row=0, sticky=NW)
        self.selected_file.grid(row=0, sticky=NW)
        load_project_btn.grid(row=1, pady=10, sticky=NW)

    def launch_project(self):
        check_file_exist_and_readable(file_path=self.selected_file.file_path)
        _ = SimbaProjectPopUp(config_path=self.selected_file.file_path)
        self.load_project_frm.destroy()
        self.main_frm.destroy()


def wait_for_internet_connection(url):
    while True:
        try:
            response = urllib.request.urlopen(url, timeout=1)
            return
        except:
            pass


class SimbaProjectPopUp(ConfigReader, PopUpMixin):
    """
    Main entry to the SimBA loaded project pop-up.
    """

    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        stdout_success(f"Loaded project {config_path}", source=self.__class__.__name__)
        simongui = Toplevel()
        simongui.minsize(1300, 800)
        simongui.wm_title("LOAD PROJECT")
        simongui.columnconfigure(0, weight=1)
        simongui.rowconfigure(0, weight=1)

        self.btn_icons = get_icons_paths()
        for k in self.btn_icons.keys():
            self.btn_icons[k]["img"] = ImageTk.PhotoImage(
                image=PIL.Image.open(
                    os.path.join(
                        os.path.dirname(__file__), self.btn_icons[k]["icon_path"]
                    )
                )
            )

        tab_parent = ttk.Notebook(hxtScrollbar(simongui))

        tab2 = ttk.Frame(tab_parent)
        tab3 = ttk.Frame(tab_parent)
        tab4 = ttk.Frame(tab_parent)
        tab5 = ttk.Frame(tab_parent)
        tab6 = ttk.Frame(tab_parent)
        tab7 = ttk.Frame(tab_parent)
        tab8 = ttk.Frame(tab_parent)
        tab9 = ttk.Frame(tab_parent)
        tab10 = ttk.Frame(tab_parent)
        tab11 = ttk.Frame(tab_parent)

        tab_parent.add(
            tab2,
            text=f"{'[ Further imports (data/video/frames) ]':20s}",
            compound="left",
            image=self.btn_icons["pose"]["img"],
        )
        tab_parent.add(
            tab3,
            text=f"{'[ Video parameters ]':20s}",
            compound="left",
            image=self.btn_icons["calipher"]["img"],
        )
        tab_parent.add(
            tab4,
            text=f"{'[ Outlier correction ]':20s}",
            compound="left",
            image=self.btn_icons["outlier"]["img"],
        )
        tab_parent.add(
            tab6,
            text=f"{'[ ROI ]':10s}",
            compound="left",
            image=self.btn_icons["roi"]["img"],
        )
        tab_parent.add(
            tab5,
            text=f"{'[ Extract features ]':20s}",
            compound="left",
            image=self.btn_icons["features"]["img"],
        )
        tab_parent.add(
            tab7,
            text=f"{'[ Label behavior] ':20s}",
            compound="left",
            image=self.btn_icons["label"]["img"],
        )
        tab_parent.add(
            tab8,
            text=f"{'[ Train machine model ]':20s}",
            compound="left",
            image=self.btn_icons["clf"]["img"],
        )
        tab_parent.add(
            tab9,
            text=f"{'[ Run machine model ]':20s}",
            compound="left",
            image=self.btn_icons["clf_2"]["img"],
        )
        tab_parent.add(
            tab10,
            text=f"{'[ Visualizations ]':20s}",
            compound="left",
            image=self.btn_icons["visualize"]["img"],
        )
        tab_parent.add(
            tab11,
            text=f"{'[ Add-ons ]':20s}",
            compound="left",
            image=self.btn_icons["add_on"]["img"],
        )

        tab_parent.grid(row=0)
        tab_parent.enable_traversal()

        import_frm = LabelFrame(tab2)
        import_frm.grid(row=0, column=0, sticky=NW)

        further_methods_frm = CreateLabelFrameWithIcon(
            parent=import_frm,
            header="FURTHER METHODS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ADDITIONAL_IMPORTS.value,
        )
        extract_frm_btn = Button(
            further_methods_frm,
            text="EXTRACT FRAMES FOR ALL VIDEOS IN SIMBA PROJECT",
            fg="blue",
            command=lambda: extract_frames_from_all_videos_in_directory(
                config_path=self.config_path, directory=self.video_dir
            ),
        )
        import_frm_dir_btn = Button(
            further_methods_frm,
            text="IMPORT FRAMES DIRECTORY TO SIMBA PROJECT",
            fg="blue",
            command=lambda: ImportFrameDirectoryPopUp(config_path=self.config_path),
        )
        add_clf_btn = Button(
            further_methods_frm,
            text="ADD CLASSIFIER TO SIMBA PROJECT",
            fg="blue",
            command=lambda: AddClfPopUp(config_path=self.config_path),
        )
        remove_clf_btn = Button(
            further_methods_frm,
            text="REMOVE CLASSIFIER FROM SIMBA PROJECT",
            fg="blue",
            command=lambda: RemoveAClassifierPopUp(config_path=self.config_path),
        )
        archive_files_btn = Button(
            further_methods_frm,
            text="ARCHIVE PROCESSED FILES IN SIMBA PROJECT",
            fg="blue",
            command=lambda: ArchiveProcessedFilesPopUp(config_path=self.config_path),
        )
        reverse_btn = Button(
            further_methods_frm,
            text="REVERSE TRACKING IDENTITIES IN SIMBA PROJECT",
            fg="blue",
            command=lambda: None,
        )
        interpolate_btn = Button(
            further_methods_frm,
            text="INTERPOLATE POSE IN SIMBA PROJECT",
            fg="blue",
            command=lambda: InterpolatePopUp(config_path=self.config_path),
        )
        smooth_btn = Button(
            further_methods_frm,
            text="SMOOTH POSE IN SIMBA PROJECT",
            fg="blue",
            command=lambda: SmoothingPopUp(config_path=self.config_path),
        )

        label_setscale = CreateLabelFrameWithIcon(
            parent=tab3,
            header="VIDEO PARAMETERS (FPS, RESOLUTION, PPX/MM ....)",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_PARAMETERS.value,
        )
        # label_setscale = LabelFrame(tab3,text='VIDEO PARAMETERS (FPS, RESOLUTION, PPX/MM ....)', font=Formats.LABELFRAME_HEADER_FORMAT.value, pady=5,padx=5,fg='black')
        self.distance_in_mm_eb = Entry_Box(
            label_setscale, "KNOWN DISTANCE (MILLIMETERS)", "25", validation="numeric"
        )
        button_setdistanceinmm = Button(
            label_setscale,
            text="AUTO-POPULATE",
            fg="green",
            command=lambda: self.set_distance_mm(),
        )

        button_setscale = Button(
            label_setscale,
            text="CONFIGURE VIDEO PARAMETERS",
            compound="left",
            image=self.btn_icons["calipher"]["img"],
            relief=RAISED,
            fg="blue",
            command=lambda: self.create_video_info_table(),
        )
        button_setscale.image = self.btn_icons["calipher"]["img"]

        self.new_ROI_frm = CreateLabelFrameWithIcon(
            parent=tab6,
            header="SIMBA ROI INTERFACE",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI.value,
        )
        self.start_new_ROI = Button(
            self.new_ROI_frm,
            text="DEFINE ROIs",
            fg="green",
            compound="left",
            image=self.btn_icons["roi"]["img"],
            relief=RAISED,
            command=lambda: ROI_menu(self.config_path),
        )
        self.start_new_ROI.image = self.btn_icons["roi"]["img"]

        self.delete_all_ROIs = Button(
            self.new_ROI_frm,
            text="DELETE ALL ROI DEFINITIONS",
            fg="red",
            compound="left",
            image=self.btn_icons["trash"]["img"],
            command=lambda: delete_all_ROIs(self.config_path),
        )
        self.delete_all_ROIs.image = self.btn_icons["trash"]["img"]

        self.standardize_roi_size_popup_btn = Button(
            self.new_ROI_frm,
            text="STANDARDIZE ROI SIZES",
            fg="blue",
            compound="left",
            image=self.btn_icons["calipher"]["img"],
            command=lambda: ROISizeStandardizerPopUp(config_path=self.config_path),
        )
        self.standardize_roi_size_popup_btn.image = self.btn_icons["calipher"]["img"]

        self.new_ROI_frm.grid(row=0, sticky=NW)
        self.start_new_ROI.grid(row=0, sticky=NW)
        self.delete_all_ROIs.grid(row=1, column=0, sticky=NW)
        self.standardize_roi_size_popup_btn.grid(row=2, column=0, sticky=NW)

        self.roi_draw = LabelFrame(
            tab6, text="ANALYZE ROI DATA", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        analyze_roi_btn = Button(
            self.roi_draw,
            text="ANALYZE ROI DATA: AGGREGATES",
            fg="green",
            command=lambda: ROIAnalysisPopUp(config_path=self.config_path),
        )
        analyze_roi_time_bins_btn = Button(
            self.roi_draw,
            text="ANALYZE ROI DATA: TIME-BINS",
            fg="blue",
            command=lambda: ROIAnalysisTimeBinsPopUp(config_path=self.config_path),
        )

        self.roi_draw.grid(row=0, column=1, sticky=N)
        analyze_roi_btn.grid(row=0, sticky="NW")
        analyze_roi_time_bins_btn.grid(row=1, sticky="NW")

        ###plot roi
        self.roi_draw1 = LabelFrame(
            tab6, text="VISUALIZE ROI DATA", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )

        # button
        visualizeROI = Button(
            self.roi_draw1,
            text="VISUALIZE ROI TRACKING",
            fg="green",
            command=lambda: VisualizeROITrackingPopUp(config_path=self.config_path),
        )
        visualizeROIfeature = Button(
            self.roi_draw1,
            text="VISUALIZE ROI FEATURES",
            fg="blue",
            command=lambda: VisualizeROIFeaturesPopUp(config_path=self.config_path),
        )

        ##organize
        self.roi_draw1.grid(row=0, column=2, sticky=N)
        visualizeROI.grid(row=0, sticky="NW")
        visualizeROIfeature.grid(row=1, sticky="NW")

        processmovementdupLabel = LabelFrame(
            tab6,
            text="OTHER ANALYSES / VISUALIZATIONS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        analyze_distances_velocity_btn = Button(
            processmovementdupLabel,
            text="ANALYZE DISTANCES / VELOCITY: AGGREGATES",
            fg="green",
            command=lambda: MovementAnalysisPopUp(config_path=self.config_path),
        )
        analyze_distances_velocity_timebins_btn = Button(
            processmovementdupLabel,
            text="ANALYZE DISTANCES / VELOCITY: TIME-BINS",
            fg="blue",
            command=lambda: MovementAnalysisTimeBinsPopUp(config_path=self.config_path),
        )

        heatmaps_location_button = Button(
            processmovementdupLabel,
            text="CREATE LOCATION HEATMAPS",
            fg="red",
            command=lambda: HeatmapLocationPopup(config_path=self.config_path),
        )

        button_lineplot = Button(
            processmovementdupLabel,
            text="CREATE PATH PLOTS",
            fg="orange",
            command=lambda: QuickLineplotPopup(config_path=self.config_path),
        )
        button_analyzeDirection = Button(
            processmovementdupLabel,
            text="ANALYZE DIRECTIONALITY BETWEEN ANIMALS",
            fg="pink",
            command=lambda: AnimalDirectingAnimalPopUp(config_path=self.config_path),
        )
        button_visualizeDirection = Button(
            processmovementdupLabel,
            text="VISUALIZE DIRECTIONALITY BETWEEN ANIMALS",
            fg="brown",
            command=lambda: DirectingOtherAnimalsVisualizerPopUp(
                config_path=self.config_path
            ),
        )
        button_analyzeDirection_bp = Button(
            processmovementdupLabel,
            text="ANALYZE DIRECTIONALITY BETWEEN BODY PARTS",
            fg="purple",
            command=lambda: DirectionAnimalToBodyPartSettingsPopUp(
                config_path=self.config_path
            ),
        )
        button_visualizeDirection_bp = Button(
            processmovementdupLabel,
            text="VISUALIZE DIRECTIONALITY BETWEEN BODY PARTS",
            fg="black",
            command=lambda: DirectingAnimalToBodyPartVisualizerPopUp(
                config_path=self.config_path
            ),
        )
        btn_agg_boolean_conditional_statistics = Button(
            processmovementdupLabel,
            text="AGGREGATE BOOLEAN CONDITIONAL STATISTICS",
            fg="grey",
            command=lambda: BooleanConditionalSlicerPopUp(config_path=self.config_path),
        )

        # organize
        processmovementdupLabel.grid(row=0, column=3, sticky=NW)
        analyze_distances_velocity_btn.grid(row=0, sticky=NW)
        heatmaps_location_button.grid(row=1, sticky=NW)
        analyze_distances_velocity_timebins_btn.grid(row=2, sticky=NW)
        button_lineplot.grid(row=3, sticky=NW)
        button_analyzeDirection.grid(row=4, sticky=NW)
        button_visualizeDirection.grid(row=5, sticky=NW)
        button_analyzeDirection_bp.grid(row=6, sticky=NW)
        button_visualizeDirection_bp.grid(row=7, sticky=NW)
        btn_agg_boolean_conditional_statistics.grid(row=8, sticky=NW)

        label_outliercorrection = CreateLabelFrameWithIcon(
            parent=tab4,
            header="OUTLIER CORRECTION",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.OUTLIERS_DOC.value,
        )
        # label_outliercorrection = LabelFrame(tab4,text='OUTLIER CORRECTION',font=Formats.LABELFRAME_HEADER_FORMAT.value,pady=5,padx=5,fg='black')
        button_settings_outlier = Button(
            label_outliercorrection,
            text="SETTINGS",
            fg="blue",
            command=lambda: OutlierSettingsPopUp(config_path=self.config_path),
        )
        button_outliercorrection = Button(
            label_outliercorrection,
            text="RUN OUTLIER CORRECTION",
            fg="green",
            command=lambda: threading.Thread(target=self.correct_outlier).start(),
        )
        button_skipOC = Button(
            label_outliercorrection,
            text="SKIP OUTLIER CORRECTION (CAUTION)",
            fg="red",
            command=lambda: threading.Thread(
                target=self.initiate_skip_outlier_correction
            ).start(),
        )

        # extract features
        label_extractfeatures = CreateLabelFrameWithIcon(
            parent=tab5,
            header="EXTRACT FEATURES",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.EXTRACT_FEATURES.value,
        )
        button_extractfeatures = Button(
            label_extractfeatures,
            text="EXTRACT FEATURES",
            fg="blue",
            command=lambda: threading.Thread(
                target=self.run_feature_extraction
            ).start(),
        )

        def activate(box, *args):
            for entry in args:
                if box.get() == 0:
                    entry.configure(state="disabled")
                elif box.get() == 1:
                    entry.configure(state="normal")

        labelframe_usrdef = LabelFrame(label_extractfeatures)
        self.scriptfile = FileSelect(
            labelframe_usrdef, "Script path", file_types=[("Python .py file", "*.py")]
        )
        self.scriptfile.btnFind.configure(state="disabled")
        self.user_defined_var = BooleanVar(value=False)
        userscript = Checkbutton(
            labelframe_usrdef,
            text="Apply user-defined feature extraction script",
            variable=self.user_defined_var,
            command=lambda: activate(self.user_defined_var, self.scriptfile.btnFind),
        )

        roi_feature_frm = CreateLabelFrameWithIcon(
            parent=tab5,
            header="APPEND ROI FEATURES",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.APPEND_ROI_FEATURES.value,
        )
        append_roi_features_by_animal = Button(
            roi_feature_frm,
            text="APPEND ROI DATA TO FEATURES: BY ANIMAL (CAUTION)",
            fg="red",
            command=lambda: AppendROIFeaturesByAnimalPopUp(
                config_path=self.config_path
            ),
        )
        append_roi_features_by_body_part = Button(
            roi_feature_frm,
            text="APPEND ROI DATA TO FEATURES: BY BODY-PARTS (CAUTION)",
            fg="orange",
            command=lambda: AppendROIFeaturesByBodyPartPopUp(
                config_path=self.config_path
            ),
        )
        remove_roi_features_from_feature_set = Button(
            roi_feature_frm,
            text="REMOVE ROI FEATURES FROM FEATURE SET",
            fg="darkred",
            command=lambda: RemoveROIFeaturesPopUp(
                config_path=self.config_path, dataset="features_extracted"
            ),
        )

        feature_tools_frm = LabelFrame(
            tab5,
            text="FEATURE TOOLS",
            pady=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        compute_feature_subset_btn = Button(
            feature_tools_frm,
            text="CALCULATE FEATURE SUBSETS",
            fg="blue",
            command=lambda: FeatureSubsetExtractorPopUp(config_path=self.config_path),
        )

        label_behavior_frm = CreateLabelFrameWithIcon(
            parent=tab7,
            header="LABEL BEHAVIOR",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.LABEL_BEHAVIOR.value,
        )
        select_video_btn_new = Button(
            label_behavior_frm,
            text="Select video (create new video annotation)",
            command=lambda: select_labelling_video(
                config_path=self.config_path,
                threshold_dict=None,
                setting="from_scratch",
                continuing=False,
            ),
        )
        select_video_btn_continue = Button(
            label_behavior_frm,
            text="Select video (continue existing video annotation)",
            command=lambda: select_labelling_video(
                config_path=self.config_path,
                threshold_dict=None,
                setting=None,
                continuing=True,
            ),
        )
        label_thirdpartyann = CreateLabelFrameWithIcon(
            parent=tab7,
            header="IMPORT THIRD-PARTY BEHAVIOR ANNOTATIONS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.THIRD_PARTY_ANNOTATION.value,
        )
        button_importmars = Button(
            label_thirdpartyann,
            text="Import MARS Annotation (select folder with .annot files)",
            fg="blue",
            command=self.importMARS,
        )
        button_importboris = Button(
            label_thirdpartyann,
            text="Import BORIS Annotation (select folder with .csv files)",
            fg="green",
            command=self.importBoris,
        )
        button_importsolomon = Button(
            label_thirdpartyann,
            text="Import SOLOMON Annotation (select folder with .csv files",
            fg="purple",
            command=self.importSolomon,
        )
        button_importethovision = Button(
            label_thirdpartyann,
            text="Import ETHOVISION Annotation (select folder with .xls/xlsx files)",
            fg="blue",
            command=self.import_ethovision,
        )
        button_importdeepethogram = Button(
            label_thirdpartyann,
            text="Import DEEPETHOGRAM Annotation (select folder with .csv files)",
            fg="green",
            command=self.import_deepethogram,
        )
        import_observer_btn = Button(
            label_thirdpartyann,
            text="Import NOLDUS OBSERVER Annotation (select folder with .xls/xlsx files)",
            fg="purple",
            command=self.import_noldus_observer,
        )

        label_pseudo = CreateLabelFrameWithIcon(
            parent=tab7,
            header="PSEUDO-LABELLING",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.PSEUDO_LBL.value,
        )
        pseudo_intructions_lbl_1 = Label(
            label_pseudo,
            text="Note that SimBA pseudo-labelling require initial machine predictions.",
        )
        pseudo_intructions_lbl_2 = Label(
            label_pseudo,
            text="Click here more information on how to use the SimBA pseudo-labelling interface.",
            cursor="hand2",
            fg="blue",
        )
        pseudo_intructions_lbl_2.bind(
            "<Button-1>",
            lambda e: self.callback(
                "https://github.com/sgoldenlab/simba/blob/master/docs/label_behavior.md"
            ),
        )
        pLabel_framedir = FileSelect(label_pseudo, "Video Path", lblwidth="10")
        plabelframe_threshold = LabelFrame(
            label_pseudo, text="Threshold", pady=5, padx=5
        )
        plabel_threshold = [0] * len(self.clf_names)
        for count, i in enumerate(list(self.clf_names)):
            plabel_threshold[count] = Entry_Box(plabelframe_threshold, str(i), "20")
            plabel_threshold[count].grid(row=count + 2, sticky=W)

        pseudo_lbl_btn = Button(
            label_pseudo,
            text="Correct labels",
            command=lambda: select_labelling_video(
                config_path=self.config_path,
                threshold_dict=dict(zip(self.clf_names, plabel_threshold)),
                setting="pseudo",
                continuing=False,
                video_file_path=pLabel_framedir.file_path,
            ),
        )

        label_adv_label = CreateLabelFrameWithIcon(
            parent=tab7,
            header="ADVANCED LABEL BEHAVIOR",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ADVANCED_LBL.value,
        )
        label_adv_note_1 = Label(
            label_adv_label,
            text="Note that you will have to specify the presence of *both* behavior and non-behavior on your own.",
        )
        label_adv_note_2 = Label(
            label_adv_label,
            text="Click here more information on how to use the SimBA labelling interface.",
            cursor="hand2",
            fg="blue",
        )
        label_adv_note_2.bind(
            "<Button-1>",
            lambda e: self.callback(
                "https://github.com/sgoldenlab/simba/blob/master/docs/advanced_labelling.md"
            ),
        )
        adv_label_btn_new = Button(
            label_adv_label,
            text="Select video (create new video annotation)",
            command=lambda: select_labelling_video_advanced(
                config_path=self.config_path, continuing=False
            ),
        )
        adv_label_btn_continue = Button(
            label_adv_label,
            text="Select video (continue existing video annotation)",
            command=lambda: select_labelling_video_advanced(
                config_path=self.config_path, continuing=True
            ),
        )

        targeted_clip_annotator_frm = CreateLabelFrameWithIcon(
            parent=tab7,
            header="TARGETED CLIP ANNOTATOR",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ADVANCED_LBL.value,
        )
        targeted_clip_annotator_note = Label(
            targeted_clip_annotator_frm,
            text="A bout annotator that creates annotated clips from videos associated with ML results.",
        )
        targeted_clip_annotator_btn = Button(
            targeted_clip_annotator_frm,
            text="Select video",
            command=lambda: select_labelling_video_targeted_clips(
                config_path=self.config_path
            ),
        )

        lbl_tools_frm = LabelFrame(
            tab7,
            text="LABELLING TOOLS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        visualize_annotation_img_btn = Button(
            lbl_tools_frm,
            text="Visualize annotations",
            fg="blue",
            command=lambda: ExtractAnnotationFramesPopUp(config_path=self.config_path),
        )
        third_party_annotations_btn = Button(
            lbl_tools_frm,
            text="Append third-party annotations",
            fg="purple",
            command=lambda: ThirdPartyAnnotatorAppenderPopUp(
                config_path=self.config_path
            ),
        )
        remove_roi_features_from_annotation_set = Button(
            lbl_tools_frm,
            text="Remove ROI features from label set",
            fg="darkred",
            command=lambda: RemoveROIFeaturesPopUp(
                config_path=self.config_path, dataset="targets_inserted"
            ),
        )

        label_trainmachinemodel = CreateLabelFrameWithIcon(
            parent=tab8,
            header="TRAIN MACHINE MODELS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.TRAIN_ML_MODEL.value,
        )
        button_trainmachinesettings = Button(
            label_trainmachinemodel,
            text="SETTINGS",
            fg="darkorange",
            command=self.trainmachinemodelsetting,
        )
        button_trainmachinemodel = Button(
            label_trainmachinemodel,
            text="TRAIN SINGLE MODEL (GLOBAL ENVIRONMENT)",
            fg="blue",
            command=lambda: threading.Thread(
                target=self.train_single_model(config_path=self.config_path)
            ).start(),
        )
        button_train_multimodel = Button(
            label_trainmachinemodel,
            text="TRAIN MULTIPLE MODELS (ONE FOR EACH SAVED SETTING)",
            fg="green",
            command=lambda: threading.Thread(
                target=self.train_multiple_models_from_meta(
                    config_path=self.config_path
                )
            ).start(),
        )

        label_model_validation = CreateLabelFrameWithIcon(
            parent=tab9,
            header="VALIDATE MODEL ON SINGLE VIDEO",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.OUT_OF_SAMPLE_VALIDATION.value,
        )
        # label_model_validation = LabelFrame(tab9, text='VALIDATE MODEL ON SINGLE VIDEO', pady=5, padx=5, font=("Helvetica", 12, 'bold'), fg='blue')
        self.csvfile = FileSelect(
            label_model_validation,
            "SELECT DATA FEATURE FILE",
            color="blue",
            lblwidth=30,
            file_types=[("SimBA CSV", "*.csv"), ("SimBA PARQUET", "*.parquet")],
        )
        self.modelfile = FileSelect(
            label_model_validation, "SELECT MODEL FILE", color="blue", lblwidth=30
        )
        button_runvalidmodel = Button(
            label_model_validation,
            text="RUN MODEL",
            fg="blue",
            command=lambda: threading.Thread(
                target=self.validate_model_first_step
            ).start(),
        )

        button_generateplot = Button(
            label_model_validation,
            text="INTERACTIVE PROBABILITY PLOT",
            fg="blue",
            command=lambda: self.launch_interactive_plot(),
        )
        self.dis_threshold = Entry_Box(
            label_model_validation, "DISCRIMINATION THRESHOLD (0.0-1.0):", "30"
        )
        self.min_behaviorbout = Entry_Box(
            label_model_validation,
            "MINIMUM BOUT LENGTH (MS):",
            "30",
            validation="numeric",
        )
        button_validate_model = Button(
            label_model_validation,
            text="CREATE VALIDATION VIDEO",
            fg="blue",
            command=lambda: ValidationVideoPopUp(
                config_path=config_path, simba_main_frm=self
            ),
        )

        label_runmachinemodel = CreateLabelFrameWithIcon(
            parent=tab9,
            header="RUN MACHINE MODEL",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.SCENARIO_2.value,
        )
        button_run_rfmodelsettings = Button(
            label_runmachinemodel,
            text="MODEL SETTINGS",
            fg="green",
            compound="left",
            image=self.btn_icons["settings"]["img"],
            command=lambda: SetMachineModelParameters(config_path=self.config_path),
        )
        button_run_rfmodelsettings.image = self.btn_icons["settings"]["img"]

        button_runmachinemodel = Button(
            label_runmachinemodel,
            text="RUN MODELS",
            fg="green",
            compound="left",
            image=self.btn_icons["clf"]["img"],
            command=lambda: threading.Thread(target=self.runrfmodel).start(),
        )
        button_runmachinemodel.image = self.btn_icons["clf"]["img"]

        kleinberg_button = Button(
            label_runmachinemodel,
            text="KLEINBERG SMOOTHING",
            fg="green",
            command=lambda: KleinbergPopUp(config_path=self.config_path),
        )
        fsttc_button = Button(
            label_runmachinemodel,
            text="FSTTC",
            fg="green",
            command=lambda: FSTTCPopUp(config_path=self.config_path),
        )
        mutual_exclusivity = Button(
            label_runmachinemodel,
            text="MUTUAL EXCLUSIVITY CORRECTION",
            fg="green",
            command=lambda: MutualExclusivityPupUp(config_path=self.config_path),
        )

        label_machineresults = CreateLabelFrameWithIcon(
            parent=tab9,
            header="ANALYZE MACHINE RESULTS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ANALYZE_ML_RESULTS.value,
        )
        button_process_datalog = Button(
            label_machineresults,
            text="ANALYZE MACHINE PREDICTIONS: AGGREGATES",
            fg="blue",
            command=lambda: ClfDescriptiveStatsPopUp(config_path=self.config_path),
        )
        button_process_movement = Button(
            label_machineresults,
            text="ANALYZE DISTANCES/VELOCITY: AGGREGATES",
            fg="blue",
            command=lambda: MovementAnalysisPopUp(config_path=self.config_path),
        )
        button_movebins = Button(
            label_machineresults,
            text="ANALYZE DISTANCES/VELOCITY: TIME BINS",
            fg="blue",
            command=lambda: MovementAnalysisTimeBinsPopUp(config_path=self.config_path),
        )
        button_classifierbins = Button(
            label_machineresults,
            text="ANALYZE MACHINE PREDICTIONS: TIME-BINS",
            fg="blue",
            command=lambda: TimeBinsClfPopUp(config_path=self.config_path),
        )
        button_classifier_ROI = Button(
            label_machineresults,
            text="ANALYZE MACHINE PREDICTION: BY ROI",
            fg="blue",
            command=lambda: ClfByROIPopUp(config_path=self.config_path),
        )
        button_severity = Button(
            label_machineresults,
            text="ANALYZE MACHINE PREDICTION: BY SEVERITY",
            fg="blue",
            command=lambda: AnalyzeSeverityPopUp(config_path=self.config_path),
        )

        visualization_frm = CreateLabelFrameWithIcon(
            parent=tab10,
            header="DATA VISUALIZATIONS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VISUALIZATION.value,
        )
        sklearn_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE CLASSIFICATIONS",
            fg="black",
            command=lambda: SklearnVisualizationPopUp(config_path=self.config_path),
        )
        sklearn_visualization_btn.grid(row=0, column=0, sticky=NW)
        gantt_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE GANTT",
            fg="blue",
            command=lambda: GanttPlotPopUp(config_path=self.config_path),
        )
        gantt_visualization_btn.grid(row=1, column=0, sticky=NW)
        probability_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE PROBABILITIES",
            fg="green",
            command=lambda: VisualizeClassificationProbabilityPopUp(
                config_path=self.config_path
            ),
        )
        probability_visualization_btn.grid(row=2, column=0, sticky=NW)
        path_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE PATHS",
            fg="orange",
            command=lambda: PathPlotPopUp(config_path=self.config_path),
        )
        path_visualization_btn.grid(row=3, column=0, sticky=NW)
        distance_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE DISTANCES",
            fg="red",
            command=lambda: DistancePlotterPopUp(config_path=self.config_path),
        )
        distance_visualization_btn.grid(row=4, column=0, sticky=NW)
        heatmap_clf_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE CLASSIFICATION HEATMAPS",
            fg="pink",
            command=lambda: HeatmapClfPopUp(config_path=self.config_path),
        )
        heatmap_clf_visualization_btn.grid(row=5, column=0, sticky=NW)
        data_plot_visualization_btn = Button(
            visualization_frm,
            text="VISUALIZE DATA PLOTS",
            fg="purple",
            command=lambda: DataPlotterPopUp(config_path=self.config_path),
        )
        data_plot_visualization_btn.grid(row=6, column=0, sticky=NW)
        clf_validation_btn = Button(
            visualization_frm,
            text="CLASSIFIER VALIDATION CLIPS",
            fg="blue",
            command=lambda: ClassifierValidationPopUp(config_path=self.config_path),
        )
        clf_validation_btn.grid(row=7, column=0, sticky=NW)
        merge_frm = CreateLabelFrameWithIcon(
            parent=tab10,
            header="MERGE FRAMES",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.CONCAT_VIDEOS.value,
        )
        merge_frm_btn = Button(
            merge_frm,
            text="MERGE FRAMES",
            fg="black",
            command=lambda: ConcatenatorPopUp(config_path=self.config_path),
        )
        plotlyInterface = CreateLabelFrameWithIcon(
            parent=tab10,
            header="PLOTLY / DASH",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.PLOTLY.value,
        )
        plotlyInterfaceTitles = [
            "Sklearn results",
            "Time bin analyses",
            "Probabilities",
            "Severity analysis",
        ]
        toIncludeVar = []
        for i in range(len(plotlyInterfaceTitles) + 1):
            toIncludeVar.append(IntVar())
        plotlyCheckbox = [0] * (len(plotlyInterfaceTitles) + 1)
        for i in range(len(plotlyInterfaceTitles)):
            plotlyCheckbox[i] = Checkbutton(
                plotlyInterface, text=plotlyInterfaceTitles[i], variable=toIncludeVar[i]
            )
            plotlyCheckbox[i].grid(row=i, sticky=W)

        button_save_plotly_file = Button(
            plotlyInterface,
            text="Save SimBA / Plotly dataset",
            command=lambda: self.generateSimBPlotlyFile(toIncludeVar),
        )
        self.plotly_file = FileSelect(
            plotlyInterface,
            "SimBA Dashboard file (H5)",
            title="Select SimBA/Plotly dataset (h5)",
        )
        self.groups_file = FileSelect(
            plotlyInterface, "SimBA Groups file (CSV)", title="Select groups file (csv"
        )
        button_open_plotly_interface = Button(
            plotlyInterface,
            text="Open SimBA / Plotly dataset",
            fg="black",
            command=lambda: [self.open_plotly_interface("http://127.0.0.1:8050")],
        )

        # addons
        lbl_addon = LabelFrame(
            tab11,
            text="SimBA Expansions",
            pady=5,
            padx=5,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        button_bel = Button(
            lbl_addon,
            text="Pup retrieval - Analysis Protocol 1",
            fg="blue",
            command=lambda: PupRetrievalPopUp(config_path=self.config_path),
        )
        cue_light_analyser_btn = Button(
            lbl_addon,
            text="Cue light analysis",
            fg="red",
            command=lambda: CueLightAnalyzerMenu(config_path=self.config_path),
        )
        anchored_roi_analysis_btn = Button(
            lbl_addon,
            text="Animal-anchored ROI analysis",
            fg="orange",
            command=lambda: BoundaryMenus(config_path=self.config_path),
        )
        # unsupervised_btn = Button(lbl_addon,text='Unsupervised analysis', fg='purple', command=lambda: UnsupervisedGUI(config_path=self.config_path))

        self.create_import_videos_menu(parent_frm=import_frm, idx_row=0, idx_column=0)
        self.create_import_pose_menu(parent_frm=import_frm, idx_row=1, idx_column=0)
        further_methods_frm.grid(row=0, column=1, sticky=NW, pady=5, padx=5)
        extract_frm_btn.grid(row=1, column=0, sticky=NW)
        import_frm_dir_btn.grid(row=2, column=0, sticky=NW)
        add_clf_btn.grid(row=3, column=0, sticky=NW)
        remove_clf_btn.grid(row=4, column=0, sticky=NW)
        archive_files_btn.grid(row=5, column=0, sticky=NW)
        reverse_btn.grid(row=6, column=0, sticky=NW)
        interpolate_btn.grid(row=7, column=0, sticky=NW)
        smooth_btn.grid(row=8, column=0, sticky=NW)

        label_setscale.grid(row=0, sticky=NW, pady=5, padx=5)
        self.distance_in_mm_eb.grid(row=0, column=0, sticky=NW)
        button_setdistanceinmm.grid(row=0, column=1, sticky=NW)
        button_setscale.grid(row=1, column=0, sticky=NW)

        label_outliercorrection.grid(row=0, sticky=W)
        button_settings_outlier.grid(row=0, sticky=W)
        button_outliercorrection.grid(row=1, sticky=W)
        button_skipOC.grid(row=2, sticky=W, pady=5)

        label_extractfeatures.grid(row=0, column=0, sticky=NW)
        button_extractfeatures.grid(row=1, column=0, sticky=NW)
        labelframe_usrdef.grid(row=0, column=0, sticky=NW, pady=15)
        userscript.grid(row=1, column=0, sticky=NW)
        self.scriptfile.grid(row=2, column=0, sticky=NW)

        roi_feature_frm.grid(row=1, column=0, sticky=NW)
        append_roi_features_by_animal.grid(row=0, column=0, sticky=NW)
        append_roi_features_by_body_part.grid(row=1, column=0, sticky=NW)
        remove_roi_features_from_feature_set.grid(row=2, column=0, sticky=NW)

        feature_tools_frm.grid(row=2, column=0, sticky=NW)
        compute_feature_subset_btn.grid(row=0, column=0, sticky=NW)

        label_behavior_frm.grid(row=5, sticky=W)
        select_video_btn_new.grid(row=0, sticky=W)
        select_video_btn_continue.grid(row=1, sticky=W)

        label_pseudo.grid(row=6, sticky=W, pady=10)

        pLabel_framedir.grid(row=0, sticky=W)
        pseudo_intructions_lbl_1.grid(row=1, sticky=W)
        pseudo_intructions_lbl_2.grid(row=2, sticky=W)
        plabelframe_threshold.grid(row=4, sticky=W)
        pseudo_lbl_btn.grid(row=5, sticky=W)

        label_adv_label.grid(row=7, column=0, sticky=NW)
        label_adv_note_1.grid(row=0, column=0, sticky=NW)
        label_adv_note_2.grid(row=1, column=0, sticky=NW)
        adv_label_btn_new.grid(row=3, column=0, sticky=NW)
        adv_label_btn_continue.grid(row=4, column=0, sticky=NW)

        targeted_clip_annotator_frm.grid(row=8, column=0, sticky=NW)
        targeted_clip_annotator_note.grid(row=0, column=0, sticky=NW)
        targeted_clip_annotator_btn.grid(row=1, column=0, sticky=NW)

        label_thirdpartyann.grid(row=9, sticky=W)
        button_importmars.grid(row=0, column=0, sticky=NW)
        button_importboris.grid(row=1, column=0, sticky=NW)
        button_importsolomon.grid(row=2, column=0, sticky=NW)
        button_importethovision.grid(row=0, column=1, sticky=NW)
        button_importdeepethogram.grid(row=1, column=1, sticky=NW)
        import_observer_btn.grid(row=2, column=1, sticky=NW)

        lbl_tools_frm.grid(row=10, column=0, sticky=NW)
        visualize_annotation_img_btn.grid(row=0, column=0, sticky=NW)
        third_party_annotations_btn.grid(row=1, column=0, sticky=NW)
        remove_roi_features_from_annotation_set.grid(row=2, column=0, sticky=NW)

        label_trainmachinemodel.grid(row=6, sticky=W)
        button_trainmachinesettings.grid(row=0, column=0, sticky=NW, padx=5)
        button_trainmachinemodel.grid(row=1, column=0, sticky=NW, padx=5)
        button_train_multimodel.grid(row=2, column=0, sticky=NW, padx=5)

        label_model_validation.grid(row=7, sticky=W, pady=5)
        self.csvfile.grid(row=0, sticky=W)
        self.modelfile.grid(row=1, sticky=W)
        button_runvalidmodel.grid(row=2, sticky=W)
        button_generateplot.grid(row=3, sticky=W)
        self.dis_threshold.grid(row=4, sticky=W)
        self.min_behaviorbout.grid(row=5, sticky=W)
        button_validate_model.grid(row=6, sticky=W)

        label_runmachinemodel.grid(row=8, sticky=NW)
        button_run_rfmodelsettings.grid(row=0, sticky=NW)
        button_runmachinemodel.grid(row=1, sticky=NW)
        kleinberg_button.grid(row=2, sticky=NW)
        fsttc_button.grid(row=3, sticky=NW)
        mutual_exclusivity.grid(row=4, sticky=NW)

        label_machineresults.grid(row=9, sticky=W, pady=5)
        button_process_datalog.grid(row=2, column=0, sticky=W, padx=3)
        button_process_movement.grid(row=2, column=1, sticky=W, padx=3)
        button_movebins.grid(row=3, column=1, sticky=W, padx=3)
        button_classifierbins.grid(row=3, column=0, sticky=W, padx=3)
        button_classifier_ROI.grid(row=4, column=0, sticky=W, padx=3)
        button_severity.grid(row=4, column=1, sticky=W, padx=3)

        visualization_frm.grid(row=11, column=0, sticky=W + N, padx=5)
        merge_frm.grid(row=11, column=2, sticky=W + N, padx=5)
        merge_frm_btn.grid(row=0, sticky=NW, padx=5)

        plotlyInterface.grid(row=11, column=3, sticky=W + N, padx=5)
        button_save_plotly_file.grid(row=10, sticky=W)
        self.plotly_file.grid(row=11, sticky=W)
        self.groups_file.grid(row=12, sticky=W)
        button_open_plotly_interface.grid(row=13, sticky=W)

        lbl_addon.grid(row=15, sticky=W)
        button_bel.grid(row=0, sticky=W)
        cue_light_analyser_btn.grid(row=1, sticky=NW)
        anchored_roi_analysis_btn.grid(row=2, sticky=NW)
        # unsupervised_btn.grid(row=3, sticky=NW)

    def create_video_info_table(self):
        video_info_tabler = VideoInfoTable(config_path=self.config_path)
        video_info_tabler.create_window()

    def initiate_skip_outlier_correction(self):
        outlier_correction_skipper = OutlierCorrectionSkipper(
            config_path=self.config_path
        )
        outlier_correction_skipper.run()

    def validate_model_first_step(self):
        _ = InferenceValidation(
            config_path=self.config_path,
            input_file_path=self.csvfile.file_path,
            clf_path=self.modelfile.file_path,
        )

    def train_single_model(self, config_path=None):
        model_trainer = TrainRandomForestClassifier(config_path=config_path)
        model_trainer.perform_sampling()
        model_trainer.train_model()
        model_trainer.save_model()

    def train_multiple_models_from_meta(self, config_path=None):
        model_trainer = GridSearchRandomForestClassifier(config_path=config_path)
        model_trainer.run()

    def importBoris(self):
        ann_folder = askdirectory()
        boris_appender = BorisAppender(
            config_path=self.config_path, data_dir=ann_folder
        )
        boris_appender.create_boris_master_file()
        threading.Thread(target=boris_appender.run).start()

    def importSolomon(self):
        ann_folder = askdirectory()
        solomon_importer = SolomonImporter(
            config_path=self.config_path, data_dir=ann_folder
        )
        threading.Thread(target=solomon_importer.run).start()

    def import_ethovision(self):
        ann_folder = askdirectory()
        ethovision_importer = ImportEthovision(
            config_path=self.config_path, data_dir=ann_folder
        )
        threading.Thread(target=ethovision_importer.run).start()

    def import_deepethogram(self):
        ann_folder = askdirectory()
        deepethogram_importer = DeepEthogramImporter(
            config_path=self.config_path, data_dir=ann_folder
        )
        threading.Thread(target=deepethogram_importer.run).start()

    def import_noldus_observer(self):
        directory = askdirectory()
        noldus_observer_importer = NoldusObserverImporter(
            config_path=self.config_path, data_dir=directory
        )
        threading.Thread(target=noldus_observer_importer.run).start()

    def importMARS(self):
        bento_dir = askdirectory()
        bento_appender = BentoAppender(config_path=self.config_path, data_dir=bento_dir)
        threading.Thread(target=bento_appender.run).start()

    def choose_animal_bps(self):
        if hasattr(self, "path_plot_animal_bp_frm"):
            self.path_plot_animal_bp_frm.destroy()
        self.path_plot_animal_bp_frm = LabelFrame(
            self.path_plot_frm,
            text="CHOOSE ANIMAL BODY-PARTS",
            font=("Helvetica", 10, "bold"),
            pady=5,
            padx=5,
        )
        self.path_plot_bp_dict = {}
        for animal_cnt in range(int(self.path_animal_cnt_dropdown.getChoices())):
            self.path_plot_bp_dict[animal_cnt] = DropDownMenu(
                self.path_plot_animal_bp_frm,
                "Animal {} bodypart".format(str(animal_cnt + 1)),
                self.bp_set,
                "15",
            )
            self.path_plot_bp_dict[animal_cnt].setChoices(self.bp_set[animal_cnt])
            self.path_plot_bp_dict[animal_cnt].grid(row=animal_cnt, sticky=NW)
        self.path_plot_animal_bp_frm.grid(row=2, column=0, sticky=NW)

    def launch_interactive_plot(self):
        interactive_grapher = InteractiveProbabilityGrapher(
            config_path=self.config_path,
            file_path=self.csvfile.file_path,
            model_path=self.modelfile.file_path,
        )
        interactive_grapher.run()

    def generateSimBPlotlyFile(self, var):
        inputList = []
        for i in var:
            inputList.append(i.get())
        stdout_warning(msg="SimBA plotly interface is not available.")
        pass

    def open_plotly_interface(self, url):
        try:
            self.p.kill()
            self.p2.kill()
        except:
            print("Starting plotly")
        # get h5 file path and csv file path
        filePath, groupPath = self.plotly_file.file_path, self.groups_file.file_path

        # print file read
        if filePath.endswith(".h5"):
            print("Reading in", os.path.basename(filePath))
        elif groupPath.endswith(".csv"):
            print("Reading in", os.path.basename(groupPath))

        self.p = subprocess.Popen(
            [
                sys.executable,
                os.path.join(
                    os.path.dirname(__file__), "dash_app", "SimBA_dash_app.py"
                ),
                filePath,
                groupPath,
            ]
        )
        # csvPath = os.path.join(os.path.dirname(self.config_path),'csv')
        # p = subprocess.Popen([sys.executable, r'simba\SimBA_dash_app.py', filePath, groupPath, csvPath])
        wait_for_internet_connection(url)
        self.p2 = subprocess.Popen(
            [
                sys.executable,
                os.path.join(
                    os.path.dirname(__file__), "dash_app", "run_dash_tkinter.py"
                ),
                url,
            ]
        )
        subprocess_children = [self.p, self.p2]
        atexit.register(terminate_children, subprocess_children)

    def runrfmodel(self):
        rf_model_runner = InferenceBatch(config_path=self.config_path)
        rf_model_runner.run()

    def trainmachinemodelsetting(self):
        _ = MachineModelSettingsPopUp(config_path=self.config_path)

    def run_feature_extraction(self):
        print(
            f"Pose-estimation body part setting for feature extraction: {str(self.animal_cnt)} animals {str(self.pose_setting)} body-parts"
        )
        feature_extractor_classes = get_bp_config_code_class_pairs()
        if self.user_defined_var.get():
            custom_feature_extractor = CustomFeatureExtractor(
                extractor_file_path=self.scriptfile.file_path,
                config_path=self.config_path,
            )
            custom_feature_extractor.run()
            stdout_success(
                msg="Custom feature extraction complete!",
                source=self.__class__.__name__,
            )
        else:
            if self.pose_setting not in feature_extractor_classes.keys():
                raise InvalidInputError(
                    msg=f"The project pose-configuration key is set to {self.pose_setting} which is invalid. OPTIONS: {list(feature_extractor_classes.keys())}. Check the pose-estimation setting in the project_config.ini",
                    source=self.__class__.__name__,
                )
            if self.pose_setting == "8":
                feature_extractor = feature_extractor_classes[self.pose_setting][
                    self.animal_cnt
                ](config_path=self.config_path)
            else:
                feature_extractor = feature_extractor_classes[self.pose_setting](
                    config_path=self.config_path
                )
            feature_extractor.run()

    def set_distance_mm(self):
        check_int(
            name="DISTANCE IN MILLIMETER",
            value=self.distance_in_mm_eb.entry_get,
            min_value=1,
        )
        self.config.set(
            "Frame settings", "distance_mm", self.distance_in_mm_eb.entry_get
        )
        with open(self.config_path, "w") as f:
            self.config.write(f)

    def correct_outlier(self):
        outlier_correcter_movement = OutlierCorrecterMovement(
            config_path=self.config_path
        )
        outlier_correcter_movement.run()
        outlier_correcter_location = OutlierCorrecterLocation(
            config_path=self.config_path
        )
        outlier_correcter_location.run()
        stdout_success(
            msg='Outlier corrected files located in "project_folder/csv/outlier_corrected_movement_location" directory',
            source=self.__class__.__name__,
        )

    def callback(self, url):
        webbrowser.open_new(url)


class App(object):
    def __init__(self):
        bg_path = os.path.join(os.path.dirname(__file__), Paths.BG_IMG_PATH.value)
        emojis = get_emojis()
        icon_path_windows = os.path.join(
            os.path.dirname(__file__), Paths.LOGO_ICON_WINDOWS_PATH.value
        )
        icon_path_darwin = os.path.join(
            os.path.dirname(__file__), Paths.LOGO_ICON_DARWIN_PATH.value
        )
        self.menu_icons = get_icons_paths()
        self.root = Tk()
        self.root.title("SimBA")
        self.root.minsize(750, 750)
        self.root.geometry(Formats.ROOT_WINDOW_SIZE.value)
        self.root.rowconfigure(0, weight=1)
        self.root.columnconfigure(0, weight=1)
        if currentPlatform == OS.WINDOWS.value:
            self.root.iconbitmap(icon_path_windows)
        if currentPlatform == OS.MAC.value:
            self.root.iconphoto(
                False, ImageTk.PhotoImage(PIL.Image.open(icon_path_darwin))
            )
        for k in self.menu_icons.keys():
            self.menu_icons[k]["img"] = ImageTk.PhotoImage(
                image=PIL.Image.open(
                    os.path.join(
                        os.path.dirname(__file__), self.menu_icons[k]["icon_path"]
                    )
                )
            )
        bg_img = ImageTk.PhotoImage(file=bg_path)
        background = Label(self.root, image=bg_img, bd=0, bg="white")
        background.pack(fill="both", expand=True)
        background.image = bg_img

        menu = Menu(self.root)
        self.root.config(menu=menu)

        file_menu = Menu(menu)
        menu.add_cascade(label="File", menu=file_menu)
        file_menu.add_command(
            label="Create a new project",
            compound="left",
            image=self.menu_icons["create"]["img"],
            command=lambda: ProjectCreatorPopUp(),
        )
        file_menu.add_command(
            label="Load project",
            compound="left",
            image=self.menu_icons["load"]["img"],
            command=lambda: LoadProjectPopUp(),
        )
        file_menu.add_separator()
        file_menu.add_command(
            label="Restart",
            compound="left",
            image=self.menu_icons["restart"]["img"],
            command=lambda: self.restart(),
        )
        file_menu.add_separator()
        file_menu.add_command(
            label="Exit",
            compound="left",
            image=self.menu_icons["exit"]["img"],
            command=self.root.destroy,
        )

        batch_process_menu = Menu(menu)
        menu.add_cascade(label="Process Videos", menu=batch_process_menu)
        batch_process_menu.add_command(
            label="Batch pre-process videos",
            compound="left",
            image=self.menu_icons["factory"]["img"],
            command=lambda: BatchPreProcessPopUp(),
        )

        video_process_menu = Menu(menu)
        fps_menu = Menu(video_process_menu)
        fps_menu.add_command(
            label="Change fps for single video", command=ChangeFpsSingleVideoPopUp
        )
        fps_menu.add_command(
            label="Change fps for multiple videos", command=ChangeFpsMultipleVideosPopUp
        )

        menu.add_cascade(label="Tools", menu=video_process_menu)
        video_process_menu.add_cascade(
            label="Change fps...",
            compound="left",
            image=self.menu_icons["fps"]["img"],
            menu=fps_menu,
        )
        format_menu = Menu(video_process_menu)
        format_menu.add_command(
            label="Change image file formats", command=ChangeImageFormatPopUp
        )
        format_menu.add_command(
            label="Change video file formats", command=ConvertVideoPopUp
        )
        video_process_menu.add_cascade(
            label="Change formats...",
            compound="left",
            image=self.menu_icons["convert"]["img"],
            menu=format_menu,
        )
        video_process_menu.add_command(
            label="CLAHE enhance video",
            compound="left",
            image=self.menu_icons["clahe"]["img"],
            command=CLAHEPopUp,
        )

        video_process_menu.add_command(
            label="Clip videos",
            compound="left",
            image=self.menu_icons["clip"]["img"],
            command=ClipVideoPopUp,
        )
        video_process_menu.add_command(
            label="Clip video into multiple videos",
            compound="left",
            image=self.menu_icons["clip"]["img"],
            command=MultiShortenPopUp,
        )
        video_process_menu.add_cascade(
            label="Concatenate multiple videos",
            compound="left",
            image=self.menu_icons["concat"]["img"],
            command=lambda: ConcatenatorPopUp(config_path=None),
        )
        video_process_menu.add_cascade(
            label="Concatenate two videos",
            compound="left",
            image=self.menu_icons["concat"]["img"],
            command=ConcatenatingVideosPopUp,
        )
        video_process_menu.add_command(
            label="Convert to grayscale",
            compound="left",
            image=self.menu_icons["grey"]["img"],
            command=lambda: GreyscaleSingleVideoPopUp(),
        )
        video_process_menu.add_command(
            label="Convert ROI definitions",
            compound="left",
            image=self.menu_icons["roi"]["img"],
            command=lambda: ConvertROIDefinitionsPopUp(),
        )
        convert_data_menu = Menu(video_process_menu)
        convert_data_menu.add_command(
            label="Convert CSV to parquet", command=Csv2ParquetPopUp
        )
        convert_data_menu.add_command(
            label="Convert parquet o CSV", command=Parquet2CsvPopUp
        )
        video_process_menu.add_cascade(
            label="Convert working file type...",
            compound="left",
            image=self.menu_icons["change"]["img"],
            menu=convert_data_menu,
        )

        video_process_menu.add_command(
            label="Create path plot",
            compound="left",
            image=self.menu_icons["path"]["img"],
            command=MakePathPlotPopUp,
        )
        video_process_menu.add_command(
            label="Crop videos",
            compound="left",
            image=self.menu_icons["crop"]["img"],
            command=CropVideoPopUp,
        )
        video_process_menu.add_command(
            label="Down-sample videos",
            compound="left",
            image=self.menu_icons["sample"]["img"],
            command=DownsampleVideoPopUp,
        )
        video_process_menu.add_cascade(
            label="Drop body-parts from tracking data",
            compound="left",
            image=self.menu_icons["trash"]["img"],
            command=DropTrackingDataPopUp,
        )
        extract_frames_menu = Menu(video_process_menu)
        extract_frames_menu.add_command(
            label="Extract defined frames", command=ExtractSpecificFramesPopUp
        )
        extract_frames_menu.add_command(
            label="Extract frames", command=ExtractAllFramesPopUp
        )
        extract_frames_menu.add_command(
            label="Extract frames from seq files", command=ExtractSEQFramesPopUp
        )
        video_process_menu.add_cascade(
            label="Extract frames...",
            compound="left",
            image=self.menu_icons["frames"]["img"],
            menu=extract_frames_menu,
        )

        video_process_menu.add_command(
            label="Generate gifs",
            compound="left",
            image=self.menu_icons["gif"]["img"],
            command=CreateGIFPopUP,
        )

        video_process_menu.add_command(
            label="Get mm/ppx",
            compound="left",
            image=self.menu_icons["calipher"]["img"],
            command=CalculatePixelsPerMMInVideoPopUp,
        )
        video_process_menu.add_command(
            label="Merge frames to video",
            compound="left",
            image=self.menu_icons["merge"]["img"],
            command=MergeFrames2VideoPopUp,
        )
        video_process_menu.add_command(
            label="Multi-crop",
            compound="left",
            image=self.menu_icons["crop"]["img"],
            command=MultiCropPopUp,
        )
        video_process_menu.add_command(
            label="Print classifier info...",
            compound="left",
            image=self.menu_icons["print"]["img"],
            command=PrintModelInfoPopUp,
        )

        video_process_menu.add_cascade(
            label="Reorganize Tracking Data",
            compound="left",
            image=self.menu_icons["reorganize"]["img"],
            command=PoseReorganizerPopUp,
        )
        video_process_menu.add_command(
            label="Rotate videos",
            compound="left",
            image=self.menu_icons["rotate"]["img"],
            command=VideoRotatorPopUp,
        )

        video_process_menu.add_command(
            label="Superimpose frame numbers on video",
            compound="left",
            image=self.menu_icons["trash"]["img"],
            command=lambda: SuperImposeFrameCountPopUp(),
        )
        video_process_menu.add_command(
            label="Temporal join videos",
            compound="left",
            image=self.menu_icons["stopwatch"]["img"],
            command=VideoTemporalJoinPopUp,
        )

        video_process_menu.add_cascade(
            label="Visualize pose-estimation in folder...",
            compound="left",
            image=self.menu_icons["visualize"]["img"],
            command=VisualizePoseInFolderPopUp,
        )

        help_menu = Menu(menu)
        menu.add_cascade(label="Help", menu=help_menu)

        links_menu = Menu(help_menu)
        links_menu.add_command(
            label="Download weights",
            command=lambda: webbrowser.open_new(str(r"https://osf.io/sr3ck/")),
        )
        links_menu.add_command(
            label="Download classifiers",
            command=lambda: webbrowser.open_new(str(r"https://osf.io/kwge8/")),
        )
        links_menu.add_command(
            label="Ex. feature list",
            command=lambda: webbrowser.open_new(
                str(
                    r"https://github.com/sgoldenlab/simba/blob/master/misc/Feature_description.csv"
                )
            ),
        )
        links_menu.add_command(
            label="SimBA github",
            command=lambda: webbrowser.open_new(
                str(r"https://github.com/sgoldenlab/simba")
            ),
        )
        links_menu.add_command(
            label="Gitter Chatroom",
            command=lambda: webbrowser.open_new(
                str(r"https://gitter.im/SimBA-Resource/community")
            ),
        )
        links_menu.add_command(
            label="Install FFmpeg",
            command=lambda: webbrowser.open_new(
                str(r"https://m.wikihow.com/Install-FFmpeg-on-Windows")
            ),
        )
        links_menu.add_command(
            label="Install graphviz",
            command=lambda: webbrowser.open_new(
                str(
                    r"https://bobswift.atlassian.net/wiki/spaces/GVIZ/pages/20971549/How+to+install+Graphviz+software"
                )
            ),
        )
        help_menu.add_cascade(
            label="Links",
            menu=links_menu,
            compound="left",
            image=self.menu_icons["link"]["img"],
        )
        help_menu.add_command(
            label="About",
            compound="left",
            image=self.menu_icons["about"]["img"],
            command=AboutSimBAPopUp,
        )

        self.frame = Frame(background, bd=2, relief=SUNKEN, width=750, height=300)
        self.r_click_menu = Menu(self.root, tearoff=0)
        self.r_click_menu.add_command(
            label="Copy selection", command=lambda: self.copy_selection_to_clipboard()
        )
        self.r_click_menu.add_command(
            label="Copy all", command=lambda: self.copy_all_to_clipboard()
        )
        self.r_click_menu.add_command(
            label="Paste", command=lambda: self.paste_to_txt()
        )
        self.r_click_menu.add_separator()
        self.r_click_menu.add_command(label="Clear", command=lambda: self.clean_txt())
        y_sb = Scrollbar(self.frame, orient=VERTICAL)
        self.frame.pack(expand=True)
        self.txt = Text(
            self.frame,
            bg="white",
            insertborderwidth=2,
            height=30,
            width=100,
            yscrollcommand=y_sb,
        )
        if currentPlatform == OS.WINDOWS.value:
            self.txt.bind("<Button-3>", self.show_right_click_pop_up)
        if currentPlatform == OS.MAC.value:
            self.txt.bind("<Button-2>", self.show_right_click_pop_up)
        self.txt.tag_configure(
            TagNames.GREETING.value,
            justify="center",
            foreground="blue",
            font=("Rockwell", 16, "bold"),
        )
        self.txt.tag_configure(
            TagNames.ERROR.value,
            justify="left",
            foreground="red",
            font=Formats.TKINTER_FONT.value,
        )
        self.txt.tag_configure(
            TagNames.STANDARD.value,
            justify="left",
            foreground="black",
            font=Formats.TKINTER_FONT.value,
        )
        self.txt.tag_configure(
            TagNames.COMPLETE.value,
            justify="left",
            foreground="darkgreen",
            font=Formats.TKINTER_FONT.value,
        )
        self.txt.tag_configure(
            TagNames.WARNING.value,
            justify="left",
            foreground="darkorange",
            font=Formats.TKINTER_FONT.value,
        )
        self.txt.tag_configure(
            "TABLE",
            foreground="darkorange",
            font=("Consolas", 10),
            wrap="none",
            borderwidth=0,
        )
        self.txt.insert(
            INSERT, Defaults.WELCOME_MSG.value + emojis["relaxed"] + "\n" * 2
        )
        self.txt.tag_add(TagNames.GREETING.value, "1.0", "3.25")
        y_sb.pack(side=RIGHT, fill=Y)
        self.txt.pack(expand=True, fill="both")
        y_sb.config(command=self.txt.yview)
        self.txt.config(state=DISABLED, font=Formats.TKINTER_FONT.value)

        clear_txt_btn = Button(
            self.frame,
            text=" CLEAR",
            compound=LEFT,
            image=self.menu_icons["clean"]["img"],
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            command=lambda: self.clean_txt(),
        )
        clear_txt_btn.pack(side=BOTTOM, fill=X)
        sys.stdout = StdRedirector(self.txt)

        if OS.PYTHON_VER.value != "3.6":
            PythonVersionWarning(
                msg=f"SimBA is not extensively tested beyond python 3.6. You are using python {OS.PYTHON_VER.value}. If you encounter errors in python>3.6, please report them on GitHub or Gitter and we will fix! (links in the help toolbar)",
                source=self.__class__.__name__,
            )

        if not check_ffmpeg_available():
            FFMpegNotFoundWarning(
                msg='SimBA could not find a FFMPEG installation on computer (as evaluated by "ffmpeg" returning None). SimBA works best with FFMPEG and it is recommended to install it on your computer',
                source=self.__class__.__name__,
            )

    def restart(self):
        confirm_restart = askyesno(
            title="RESTART", message="Are you sure that you want restart SimBA?"
        )
        if confirm_restart:
            self.root.destroy()
            os.execl(sys.executable, sys.executable, *sys.argv)

    def clean_txt(self):
        self.txt.config(state=NORMAL)
        self.txt.delete("1.0", END)

    def show_right_click_pop_up(self, event):
        try:
            self.r_click_menu.tk_popup(event.x_root, event.y_root)
        finally:
            self.r_click_menu.grab_release()

    def copy_selection_to_clipboard(self):
        self.root.clipboard_clear()
        text = self.txt.get("sel.first", "sel.last")
        text = text.encode("ascii", "ignore").decode()
        self.root.clipboard_append(text)

    def copy_all_to_clipboard(self):
        self.root.clipboard_clear()
        self.root.clipboard_append(self.txt.get("1.0", "end-1c"))

    def paste_to_txt(self):
        try:
            print(self.root.clipboard_get())
        except UnicodeDecodeError:
            raise InvalidInputError(
                msg="Can only paste utf-8 compatible text",
                source=self.__class__.__name__,
            )


class StdRedirector(object):
    def __init__(self, text_widget):
        self.text_space = text_widget
        self.emojis = get_emojis()

    def write(self, s: str):
        tag_name = TagNames.STANDARD.value
        try:
            s, tag_name = s.split(Defaults.STR_SPLIT_DELIMITER.value, 2)
        except ValueError:
            pass
        if (tag_name != TagNames.STANDARD.value) and (tag_name != "TABLE"):
            s = s + " " + self.emojis[tag_name]
        self.text_space.config(state=NORMAL)
        self.text_space.insert("end", s, (tag_name))
        self.text_space.update()
        self.text_space.see("end")
        self.text_space.config(state=DISABLED)

    def flush(self):
        pass


class SplashStatic:
    def __init__(self, parent):
        self.parent = parent
        self.load_splash_img()
        self.display_splash()

    def load_splash_img(self):
        splash_path_win = os.path.join(
            os.path.dirname(__file__), Paths.SPLASH_PATH_WINDOWS.value
        )
        splash_path_linux = os.path.join(
            os.path.dirname(__file__), Paths.SPLASH_PATH_LINUX.value
        )
        if currentPlatform == OS.WINDOWS.value:
            self.splash_img = PIL.Image.open(splash_path_win)
        else:
            if os.path.isfile(splash_path_linux):
                self.splash_img = PIL.Image.open(splash_path_linux)
            else:
                self.splash_img = PIL.Image.open(splash_path_win)
        self.splash_img_tk = ImageTk.PhotoImage(self.splash_img)

    def display_splash(self):
        width, height = self.splash_img.size
        half_width = int((self.parent.winfo_screenwidth() - width) // 2)
        half_height = int((self.parent.winfo_screenheight() - height) // 2)
        self.parent.geometry("%ix%i+%i+%i" % (width, height, half_width, half_height))
        Label(self.parent, image=self.splash_img_tk).pack()


class SplashMovie:
    def __init__(self):
        self.parent, self.img_cnt = Tk(), 0
        self.parent.overrideredirect(True)
        self.parent.configure(bg="white")
        splash_path = os.path.join(
            os.path.dirname(__file__), Paths.SPLASH_PATH_MOVIE.value
        )
        self.meta_ = get_video_meta_data(splash_path)
        self.cap = cv2.VideoCapture(splash_path)
        width, height = self.meta_["width"], self.meta_["height"]
        half_width = int((self.parent.winfo_screenwidth() - width) // 2)
        half_height = int((self.parent.winfo_screenheight() - height) // 2)
        self.parent.geometry("%ix%i+%i+%i" % (width, height, half_width, half_height))
        self.img_lbl = Label(self.parent, bg="white", image="")
        self.img_lbl.pack()
        self.show_animation()

    def show_animation(self):
        for frm_cnt in range(self.meta_["frame_count"] - 1):
            self.cap.set(1, frm_cnt)
            ret, frame = self.cap.read()
            if not ret:
                break
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGBA)
            frame = ImageTk.PhotoImage(image=PIL.Image.fromarray(frame))
            self.img_lbl.configure(image=frame)
            self.img_lbl.imgtk = frame
            self.parent.update()
            cv2.waitKey(max(33, int(self.meta_["fps"] / 1000)))
        self.parent.destroy()


def terminate_children(children):
    for process in children:
        process.terminate()


def main():
    if currentPlatform == OS.WINDOWS.value:
        import ctypes

        myappid = "SimBA development wheel"
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    SplashMovie()
    app = App()
    app.root.mainloop()



from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import FolderSelect
from simba.utils.read_write import (convert_csv_to_parquet,
                                    convert_parquet_to_csv)


class Csv2ParquetPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="Convert CSV directory to parquet", size=(300, 300)
        )
        frm = LabelFrame(
            self.main_frm, text="Select CSV directory", padx=5, pady=5, font="bold"
        )
        folder_path = FolderSelect(frm, "CSV folder path", title=" Select CSV folder")
        run_btn = Button(
            frm,
            text="Convert CSV to parquet",
            command=lambda: convert_csv_to_parquet(directory=folder_path.folder_path),
        )
        frm.grid(row=1, sticky=W, pady=10)
        folder_path.grid(row=0, sticky=W)
        run_btn.grid(row=1, sticky=W, pady=10)


class Parquet2CsvPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="Convert parquet directory to CSV", size=(300, 300)
        )
        frm = LabelFrame(
            self.main_frm, text="Select parquet directory", padx=5, pady=5, font="bold"
        )
        folder_path = FolderSelect(
            frm, "Parquet folder path", title=" Select parquet folder"
        )
        run_btn = Button(
            frm,
            text="Convert parquet to CSV",
            command=lambda: convert_parquet_to_csv(directory=folder_path.folder_path),
        )
        frm.grid(row=1, sticky=W, pady=10)
        folder_path.grid(row=0, sticky=W)
        run_btn.grid(row=1, sticky=W, pady=10)



__author__ = "Simon Nilsson"

import csv
import os
import platform
from configparser import ConfigParser
from typing import List

import simba
from simba.utils.enums import ConfigKey, DirNames, Dtypes, Paths
from simba.utils.errors import DirectoryExistError
from simba.utils.printing import SimbaTimer, stdout_success


class ProjectConfigCreator(object):
    """
    Create SimBA project directory tree and associated project_config.ini config file.

    :parameter str project_path: path to directory where to save the SimBA project directory tree
    :parameter str project_name: Name of the SimBA project
    :parameter List[str] target_list: Classifier names in the SimBA project
    :parameter str pose_estimation_bp_cnt: String representing the number of body-parts in the pose-estimation data used in the simba project.
                                           E.g., '4', '7', '8', '9', '14', '16' or 'user_defined', '3D_user_defined'.
    :parameter int body_part_config_idx: The index of the SimBA GUI dropdown pose-estimation selection. E.g., ``1``. I.e., the row representing
                                         your pose-estimated body-parts in `this file <https://github.com/sgoldenlab/simba/blob/master/simba/pose_configurations/bp_names/bp_names.csv>`_.
    :parameter int animal_cnt: Number of animals tracked in the input pose-estimation data.
    :parameter str file_type: The SimBA project file type. OPTIONS: ``csv`` or ``parquet``.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#part-1-create-a-new-project-1>`__.

    Examples
    ----------
    >>> _ = ProjectConfigCreator(project_path = 'project/path', project_name='project_name', target_list=['Attack'], pose_estimation_bp_cnt='16', body_part_config_idx=9, animal_cnt=2, file_type='csv')

    """

    def __init__(
        self,
        project_path: str,
        project_name: str,
        target_list: List[str],
        pose_estimation_bp_cnt: str,
        body_part_config_idx: int,
        animal_cnt: int,
        file_type: str = "csv",
    ):
        self.simba_dir = os.path.dirname(simba.__file__)
        self.animal_cnt = animal_cnt
        self.os_platform = platform.system()
        self.project_path = project_path
        self.project_name = project_name
        self.target_list = target_list
        self.pose_estimation_bp_cnt = pose_estimation_bp_cnt
        self.body_part_config_idx = body_part_config_idx
        self.file_type = file_type
        self.timer = SimbaTimer(start=True)
        self.__create_directories()
        self.__create_configparser_config()

    def __create_directories(self):
        self.project_folder = os.path.join(
            self.project_path, self.project_name, DirNames.PROJECT.value
        )
        self.models_folder = os.path.join(
            self.project_path, self.project_name, DirNames.MODEL.value
        )
        self.config_folder = os.path.join(self.project_folder, DirNames.CONFIGS.value)
        self.csv_folder = os.path.join(self.project_folder, DirNames.CSV.value)
        self.frames_folder = os.path.join(self.project_folder, DirNames.FRAMES.value)
        self.logs_folder = os.path.join(self.project_folder, DirNames.LOGS.value)
        self.measures_folder = os.path.join(self.logs_folder, DirNames.MEASURES.value)
        self.pose_configs_folder = os.path.join(
            self.measures_folder, DirNames.POSE_CONFIGS.value
        )
        self.bp_names_folder = os.path.join(
            self.pose_configs_folder, DirNames.BP_NAMES.value
        )
        self.videos_folder = os.path.join(self.project_folder, DirNames.VIDEOS.value)
        self.features_extracted_folder = os.path.join(
            self.csv_folder, DirNames.FEATURES_EXTRACTED.value
        )
        self.input_csv_folder = os.path.join(self.csv_folder, DirNames.INPUT_CSV.value)
        self.machine_results_folder = os.path.join(
            self.csv_folder, DirNames.MACHINE_RESULTS.value
        )
        self.outlier_corrected_movement_folder = os.path.join(
            self.csv_folder, DirNames.OUTLIER_MOVEMENT.value
        )
        self.outlier_corrected_location_folder = os.path.join(
            self.csv_folder, DirNames.OUTLIER_MOVEMENT_LOCATION.value
        )
        self.targets_inserted_folder = os.path.join(
            self.csv_folder, DirNames.TARGETS_INSERTED.value
        )
        self.input_folder = os.path.join(self.frames_folder, DirNames.INPUT.value)
        self.output_folder = os.path.join(self.frames_folder, DirNames.OUTPUT.value)

        folder_lst = [
            self.project_folder,
            self.models_folder,
            self.config_folder,
            self.csv_folder,
            self.frames_folder,
            self.logs_folder,
            self.videos_folder,
            self.features_extracted_folder,
            self.input_csv_folder,
            self.machine_results_folder,
            self.outlier_corrected_movement_folder,
            self.outlier_corrected_location_folder,
            self.targets_inserted_folder,
            self.input_folder,
            self.output_folder,
            self.measures_folder,
            self.pose_configs_folder,
            self.bp_names_folder,
        ]

        for folder_path in folder_lst:
            if os.path.isdir(folder_path):
                raise DirectoryExistError(
                    msg=f"SimBA tried to create {folder_path}, but it already exists. Please create your SimBA project in a new path, or move/delete your previous SimBA project"
                )
            else:
                os.makedirs(folder_path)

    def __create_configparser_config(self):
        self.config = ConfigParser(allow_no_value=True)
        self.config.add_section(ConfigKey.GENERAL_SETTINGS.value)
        self.config[ConfigKey.GENERAL_SETTINGS.value][
            ConfigKey.PROJECT_PATH.value
        ] = self.project_folder
        self.config[ConfigKey.GENERAL_SETTINGS.value][
            ConfigKey.PROJECT_NAME.value
        ] = self.project_name
        self.config[ConfigKey.GENERAL_SETTINGS.value][
            ConfigKey.FILE_TYPE.value
        ] = self.file_type
        self.config[ConfigKey.GENERAL_SETTINGS.value][ConfigKey.ANIMAL_CNT.value] = str(
            self.animal_cnt
        )
        self.config[ConfigKey.GENERAL_SETTINGS.value][
            ConfigKey.OS.value
        ] = self.os_platform

        self.config.add_section(ConfigKey.SML_SETTINGS.value)
        self.config[ConfigKey.SML_SETTINGS.value][
            ConfigKey.MODEL_DIR.value
        ] = self.models_folder
        for clf_cnt in range(len(self.target_list)):
            self.config[ConfigKey.SML_SETTINGS.value][
                "model_path_{}".format(str(clf_cnt + 1))
            ] = os.path.join(
                self.models_folder, str(self.target_list[clf_cnt]) + ".sav"
            )

        self.config[ConfigKey.SML_SETTINGS.value][ConfigKey.TARGET_CNT.value] = str(
            len(self.target_list)
        )
        for clf_cnt in range(len(self.target_list)):
            self.config[ConfigKey.SML_SETTINGS.value][
                "target_name_{}".format(str(clf_cnt + 1))
            ] = str(self.target_list[clf_cnt])

        self.config.add_section(ConfigKey.THRESHOLD_SETTINGS.value)
        for clf_cnt in range(len(self.target_list)):
            self.config[ConfigKey.THRESHOLD_SETTINGS.value][
                "threshold_{}".format(str(clf_cnt + 1))
            ] = Dtypes.NONE.value
        self.config[ConfigKey.THRESHOLD_SETTINGS.value][
            ConfigKey.SKLEARN_BP_PROB_THRESH.value
        ] = str(0.00)

        self.config.add_section(ConfigKey.MIN_BOUT_LENGTH.value)
        for clf_cnt in range(len(self.target_list)):
            self.config[ConfigKey.MIN_BOUT_LENGTH.value][
                "min_bout_{}".format(str(clf_cnt + 1))
            ] = Dtypes.NONE.value

        self.config.add_section(ConfigKey.FRAME_SETTINGS.value)
        self.config[ConfigKey.FRAME_SETTINGS.value][ConfigKey.DISTANCE_MM.value] = 0.00
        self.config.add_section(ConfigKey.LINE_PLOT_SETTINGS.value)
        self.config.add_section(ConfigKey.PATH_PLOT_SETTINGS.value)
        self.config.add_section(ConfigKey.ROI_SETTINGS.value)
        self.config.add_section(ConfigKey.DIRECTIONALITY_SETTINGS.value)
        self.config.add_section(ConfigKey.PROCESS_MOVEMENT_SETTINGS.value)

        self.config.add_section(ConfigKey.CREATE_ENSEMBLE_SETTINGS.value)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.POSE_SETTING.value
        ] = str(self.pose_estimation_bp_cnt)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.CLASSIFIER.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.TT_SIZE.value
        ] = str(0.20)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.UNDERSAMPLE_SETTING.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.UNDERSAMPLE_RATIO.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.OVERSAMPLE_SETTING.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.OVERSAMPLE_RATIO.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.RF_ESTIMATORS.value
        ] = str(2000)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.MIN_LEAF.value
        ] = str(1)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.RF_MAX_FEATURES.value
        ] = Dtypes.SQRT.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.RF_JOBS.value
        ] = str(-1)
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.RF_CRITERION.value
        ] = Dtypes.ENTROPY.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.RF_METADATA.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.EX_DECISION_TREE.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.EX_DECISION_TREE_FANCY.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.IMPORTANCE_LOG.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.IMPORTANCE_BAR_CHART.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.PERMUTATION_IMPORTANCE.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.LEARNING_CURVE.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.PRECISION_RECALL.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.IMPORTANCE_BARS_N.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.LEARNING_CURVE_K_SPLITS.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.CREATE_ENSEMBLE_SETTINGS.value][
            ConfigKey.LEARNING_DATA_SPLITS.value
        ] = Dtypes.NONE.value

        self.config.add_section(ConfigKey.MULTI_ANIMAL_ID_SETTING.value)
        self.config[ConfigKey.MULTI_ANIMAL_ID_SETTING.value][
            ConfigKey.MULTI_ANIMAL_IDS.value
        ] = Dtypes.NONE.value

        self.config.add_section(ConfigKey.OUTLIER_SETTINGS.value)
        self.config[ConfigKey.OUTLIER_SETTINGS.value][
            ConfigKey.MOVEMENT_CRITERION.value
        ] = Dtypes.NONE.value
        self.config[ConfigKey.OUTLIER_SETTINGS.value][
            ConfigKey.LOCATION_CRITERION.value
        ] = Dtypes.NONE.value

        self.config_path = os.path.join(self.project_folder, "project_config.ini")
        with open(self.config_path, "w") as file:
            self.config.write(file)

        bp_dir_path = os.path.join(self.simba_dir, Paths.SIMBA_BP_CONFIG_PATH.value)
        with open(bp_dir_path, "r", encoding="utf8") as f:
            cr = csv.reader(f, delimiter=",")  # , is default
            rows = list(cr)  # create a list of rows for instance

        chosen_bps = rows[self.body_part_config_idx]
        chosen_bps = list(filter(None, chosen_bps))

        project_bp_file_path = os.path.join(
            self.bp_names_folder, "project_bp_names.csv"
        )
        f = open(project_bp_file_path, "w+")
        for i in chosen_bps:
            f.write(i + "\n")
        f.close()
        self.timer.stop_timer()

        stdout_success(
            msg=f"Project directory tree and project_config.ini created in {self.project_folder}",
            elapsed_time=self.timer.elapsed_time_str,
        )



import datetime
import glob
import json
import os
import re
from tkinter import *
from typing import Union

import cv2

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_string_value_is_valid_video_timestamp,
                                check_int, check_nvidea_gpu_available,
                                check_that_hhmmss_start_is_before_end)
from simba.utils.enums import Keys, Links, Options
from simba.utils.errors import FFMPEGCodecGPUError, NoFilesFoundError
from simba.utils.lookups import (percent_to_crf_lookup,
                                 video_quality_to_preset_lookup)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (
    check_if_hhmmss_timestamp_is_valid_part_of_video, get_fn_ext,
    get_video_meta_data)
from simba.video_processors.batch_process_create_ffmpeg_commands import \
    FFMPEGCommandCreator
from simba.video_processors.roi_selector import ROISelector


class BatchProcessFrame(PopUpMixin):
    """
    Interactive GUI that collect user-inputs for batch processing videos (e.g., cropping,
    clipping etc.). User-selected output is stored in json file format within the user-defined `output_dir`

    .. note::
       `Batch pre-process tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial_process_videos.md>`__.

    :param str input_dir: Input folder path containing videos for bath processing.
    :param str output_dir: Output folder path for where to store the processed videos.


    Examples
    ----------
    >>> batch_preprocessor = BatchProcessFrame(input_dir=r'MyInputVideosDir', output_dir=r'MyOutputVideosDir')
    >>> batch_preprocessor.create_main_window()
    >>> batch_preprocessor.create_video_table_headings()
    >>> batch_preprocessor.create_video_rows()
    >>> batch_preprocessor.create_execute_btn()
    >>> batch_preprocessor.main_frm.mainloop()

    """

    def __init__(
        self, input_dir: Union[str, os.PathLike], output_dir: Union[str, os.PathLike]
    ):
        PopUpMixin.__init__(
            self, title="BATCH PRE-PROCESS VIDEOS IN SIMBA", size=(1600, 600)
        )
        self.input_dir, self.output_dir = input_dir, output_dir
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        self.videos_in_dir_dict, self.crop_dict = {}, {}
        self.get_input_files()
        self.percent_to_crf_lookup = percent_to_crf_lookup()
        self.cpu_video_quality = list(range(10, 110, 10))
        self.cpu_video_quality = [str(x) for x in self.cpu_video_quality]
        self.video_quality_to_preset_lookup = video_quality_to_preset_lookup()
        if len(list(self.videos_in_dir_dict.keys())) == 0:
            raise NoFilesFoundError(
                msg=f"The input directory {self.input_dir} contains ZERO video files in either .avi, .mp4, .mov, .flv, or m4v format",
                source=self.__class__.__name__,
            )
        self.max_char_vid_name = len(max(list(self.videos_in_dir_dict.keys()), key=len))

    def get_input_files(self):
        for file_path in glob.glob(self.input_dir + "/*"):
            lower_str_name = file_path.lower()
            if lower_str_name.endswith(Options.ALL_VIDEO_FORMAT_OPTIONS.value):
                _, video_name, ext = get_fn_ext(file_path)
                self.videos_in_dir_dict[video_name] = get_video_meta_data(
                    video_path=file_path, fps_as_int=False
                )
                self.videos_in_dir_dict[video_name]["extension"] = ext
                self.videos_in_dir_dict[video_name]["video_length"] = str(
                    datetime.timedelta(
                        seconds=int(
                            self.videos_in_dir_dict[video_name]["frame_count"]
                            / self.videos_in_dir_dict[video_name]["fps"]
                        )
                    )
                )
                self.videos_in_dir_dict[video_name]["video_length"] = (
                    "0" + self.videos_in_dir_dict[video_name]["video_length"]
                )
                self.videos_in_dir_dict[video_name]["file_path"] = file_path

    def create_main_window(self):
        self.quick_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="QUICK SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.BATCH_PREPROCESS.value,
        )
        self.clip_video_settings_frm = LabelFrame(
            self.quick_settings_frm, text="Clip Videos Settings", padx=5
        )
        self.quick_clip_start_entry_lbl = Label(
            self.clip_video_settings_frm, text="Start Time: "
        )
        self.quick_clip_start_entry_box_val = StringVar()
        self.quick_clip_start_entry_box_val.set("00:00:00")
        self.quick_clip_start_entry_box = Entry(
            self.clip_video_settings_frm,
            width=15,
            textvariable=self.quick_clip_start_entry_box_val,
        )
        self.quick_clip_end_entry_lbl = Label(
            self.clip_video_settings_frm, text="End Time: "
        )
        self.quick_clip_end_entry_box_val = StringVar()
        self.quick_clip_end_entry_box_val.set("00:00:00")
        self.quick_clip_end_entry_box = Entry(
            self.clip_video_settings_frm,
            width=15,
            textvariable=self.quick_clip_end_entry_box_val,
        )
        self.quick_clip_apply = Button(
            self.clip_video_settings_frm,
            text="Apply",
            command=lambda: self.apply_trim_to_all(),
        )
        self.quick_downsample_frm = LabelFrame(
            self.quick_settings_frm, text="Downsample Videos", padx=5
        )
        self.quick_downsample_width_lbl = Label(
            self.quick_downsample_frm, text="Width: "
        )
        self.quick_downsample_width_val = IntVar()
        self.quick_downsample_width_val.set(400)
        self.quick_downsample_width = Entry(
            self.quick_downsample_frm,
            width=15,
            textvariable=self.quick_downsample_width_val,
        )
        self.quick_downsample_height_lbl = Label(
            self.quick_downsample_frm, text="Height: "
        )
        self.quick_downsample_height_val = IntVar()
        self.quick_downsample_height_val.set(600)
        self.quick_downsample_height = Entry(
            self.quick_downsample_frm,
            width=15,
            textvariable=self.quick_downsample_height_val,
        )
        self.quick_downsample_apply = Button(
            self.quick_downsample_frm,
            text="Apply",
            command=lambda: self.apply_resolution_to_all(),
        )
        self.quick_set_fps = LabelFrame(
            self.quick_settings_frm, text="Change FPS", padx=5, pady=5
        )
        self.quick_fps_lbl = Label(self.quick_set_fps, text="FPS: ")
        self.quick_set_fps_val = DoubleVar()
        self.quick_set_fps_val.set(15.0)
        self.quick_fps_entry_box = Entry(
            self.quick_set_fps, width=15, textvariable=self.quick_set_fps_val
        )
        self.quick_set_fps_empty_row = Label(self.quick_set_fps, text=" ")
        self.quick_fps_apply = Button(
            self.quick_set_fps, text="Apply", command=lambda: self.apply_fps_to_all()
        )

        self.quick_set_quality = LabelFrame(
            self.quick_settings_frm, text="Output Video Quality", padx=5, pady=5
        )
        self.quick_set_quality_dropdown = DropDownMenu(
            self.quick_set_quality, "Video Quality % ", self.cpu_video_quality, "14"
        )
        self.quick_set_quality_dropdown.setChoices(100)
        self.quick_set_qualitys_empty_row = Label(self.quick_set_quality, text=" ")
        self.quick_set_quality_apply = Button(
            self.quick_set_quality,
            text="Apply",
            command=lambda: self.apply_quality_to_all(),
        )

        self.use_gpu_frm = LabelFrame(
            self.quick_settings_frm, text="GPU", padx=5, pady=10
        )
        self.use_gpu_var = BooleanVar(value=False)
        use_gpu_cb = Checkbutton(
            self.use_gpu_frm,
            text="Use GPU (reduced runtime)",
            variable=self.use_gpu_var,
            command=lambda: self.change_quality_options_cpu_gpu(),
        )
        use_gpu_cb.grid(row=0, column=0, sticky=NW)
        Label(self.use_gpu_frm, text=" ").grid(row=1, column=0)
        Label(self.use_gpu_frm, text=" ").grid(row=2, column=0)

        self.quick_settings_frm.grid(row=0, column=0, sticky=W, padx=10)
        self.clip_video_settings_frm.grid(row=0, column=0, sticky=W)
        self.quick_clip_start_entry_lbl.grid(row=0, column=0, sticky=W)
        self.quick_clip_start_entry_box.grid(row=0, column=1, sticky=W)
        self.quick_clip_end_entry_lbl.grid(row=1, column=0, sticky=W)
        self.quick_clip_end_entry_box.grid(row=1, column=1, sticky=W)
        self.quick_clip_apply.grid(row=2, column=0)
        self.quick_downsample_frm.grid(row=0, column=1, sticky=W)
        self.quick_downsample_width_lbl.grid(row=0, column=0, sticky=W)
        self.quick_downsample_width.grid(row=0, column=1, sticky=W)
        self.quick_downsample_height_lbl.grid(row=1, column=0, sticky=W)
        self.quick_downsample_height.grid(row=1, column=1, sticky=W)
        self.quick_downsample_apply.grid(row=2, column=0, sticky=W)
        self.quick_set_fps.grid(row=0, column=2, sticky=W)
        self.quick_fps_lbl.grid(row=0, column=0, sticky=W)
        self.quick_fps_entry_box.grid(row=0, column=1, sticky=W)
        self.quick_set_fps_empty_row.grid(row=1, column=1, sticky=W)
        self.quick_fps_apply.grid(row=2, column=0, sticky=W)
        self.quick_set_quality.grid(row=0, column=3, sticky=W)
        self.quick_set_quality_dropdown.grid(row=0, column=0, sticky=W)
        self.quick_set_qualitys_empty_row.grid(row=1, column=0, sticky=W)
        self.quick_set_quality_apply.grid(row=2, column=0, sticky=W)
        self.use_gpu_frm.grid(row=0, column=4, sticky=W)

    def inverse_all_cb_ticks(self, variable_name=None):
        if self.headings[variable_name].get():
            for video_name in self.videos.keys():
                self.videos[video_name][variable_name].set(True)
        if not self.headings[variable_name].get():
            for video_name in self.videos.keys():
                self.videos[video_name][variable_name].set(False)

    def change_quality_options_cpu_gpu(self):
        self.quick_set_quality_dropdown.destroy()
        if not self.use_gpu_var.get():
            self.quick_set_quality_dropdown = DropDownMenu(
                self.quick_set_quality, "Video Quality % ", self.cpu_video_quality, "14"
            )
            self.quick_set_quality_dropdown.setChoices(100)
        else:
            self.quick_set_quality_dropdown = DropDownMenu(
                self.quick_set_quality,
                "Video Quality % ",
                ["Low", "Medium", "High"],
                "14",
            )
            self.quick_set_quality_dropdown.setChoices("Medium")
        self.quick_set_quality_dropdown.grid(row=0, column=0, sticky=W)

        for video_cnt, video_name in enumerate(self.videos.keys()):
            self.videos[video_name]["video_quality_dropdown"].destroy()
            if not self.use_gpu_var.get():
                self.videos[video_name]["video_quality_dropdown"] = DropDownMenu(
                    self.videos_frm, "", self.cpu_video_quality, "5"
                )
                self.videos[video_name]["video_quality_dropdown"].setChoices(100)
            else:
                self.videos[video_name]["video_quality_dropdown"] = DropDownMenu(
                    self.videos_frm, "", ["Low", "Medium", "High"], "5"
                )
                self.videos[video_name]["video_quality_dropdown"].setChoices("Medium")
            self.videos[video_name]["video_quality_dropdown"].grid(
                row=video_cnt + 1, column=14, sticky=W
            )

    def apply_resolution_to_all(self):
        check_int(
            value=self.quick_downsample_width_val.get(),
            min_value=0,
            name=f"Quick set downsample WIDTH {self.quick_downsample_width_val.get()}",
        )
        check_int(
            value=self.quick_downsample_height_val.get(),
            min_value=0,
            name=f"Quick set downsample HEIGHT {self.quick_downsample_height_val.get()}",
        )
        for video_name in self.videos.keys():
            self.videos[video_name]["width_var"].set(
                self.quick_downsample_width_val.get()
            )
            self.videos[video_name]["height_var"].set(
                self.quick_downsample_height_val.get()
            )

    def apply_trim_to_all(self):
        check_if_string_value_is_valid_video_timestamp(
            value=self.quick_clip_start_entry_box.get(),
            name=f"Quick set clip START time {self.quick_clip_start_entry_box.get()}",
        )
        check_if_string_value_is_valid_video_timestamp(
            value=self.quick_clip_end_entry_box.get(),
            name=f"Quick set clip END time {self.quick_clip_start_entry_box.get()}",
        )
        check_that_hhmmss_start_is_before_end(
            start_time=self.quick_clip_start_entry_box.get(),
            end_time=self.quick_clip_end_entry_box.get(),
            name="Quick set START and END time",
        )
        for video_name in self.videos.keys():
            self.videos[video_name]["start_time_var"].set(
                self.quick_clip_start_entry_box.get()
            )
            self.videos[video_name]["end_time_var"].set(
                self.quick_clip_end_entry_box.get()
            )

    def apply_fps_to_all(self):
        check_float(
            value=self.quick_set_fps_val.get(),
            min_value=0,
            name=f"Quick set FPS setting {self.quick_set_fps_val.get()}",
        )
        for video_name in self.videos.keys():
            self.videos[video_name]["fps_var"].set(self.quick_set_fps_val.get())

    def apply_quality_to_all(self):
        for video_name in self.videos.keys():
            self.videos[video_name]["video_quality_dropdown"].setChoices(
                self.quick_set_quality_dropdown.getChoices()
            )

    def create_video_table_headings(self):
        self.headings = {}
        self.videos_frm = LabelFrame(
            self.main_frm,
            text="VIDEOS",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.headings["video_name_col_head"] = Label(
            self.videos_frm, text="Video Name", width=self.max_char_vid_name + 5
        )
        self.headings["crop_video_col_head"] = Label(
            self.videos_frm, text="Crop Video", width=8
        )
        self.headings["start_time_col_head"] = Label(
            self.videos_frm, text="Start Time", width=8
        )
        self.headings["end_time_col_head"] = Label(
            self.videos_frm, text="End Time", width=8
        )
        self.headings["video_quality_head"] = Label(
            self.videos_frm, text="Quality %", width=15
        )
        self.headings["clip_cb_var"] = BooleanVar()
        self.headings["shorten_all_videos_cbox"] = Checkbutton(
            self.videos_frm,
            text="Clip All Videos",
            variable=self.headings["clip_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(variable_name="clip_cb_var"),
        )
        self.headings["video_width_col_head"] = Label(
            self.videos_frm, text="Video Width", width=8
        )
        self.headings["video_height_col_head"] = Label(
            self.videos_frm, text="Video Height", width=8
        )
        self.headings["downsample_cb_var"] = BooleanVar()
        self.headings["downsample_all_videos_cbox"] = Checkbutton(
            self.videos_frm,
            text="Downsample All Videos",
            variable=self.headings["downsample_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(
                variable_name="downsample_cb_var"
            ),
        )
        self.headings["fps_col_head"] = Label(
            self.videos_frm, text="Video FPS", width=8
        )
        self.headings["fps_cb_var"] = BooleanVar()
        self.headings["change_fps_all_videos_cbox"] = Checkbutton(
            self.videos_frm,
            text="Change FPS",
            variable=self.headings["fps_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(variable_name="fps_cb_var"),
        )
        self.headings["grayscale_cb_var"] = BooleanVar()
        self.headings["grayscale_cbox"] = Checkbutton(
            self.videos_frm,
            text="Apply Greyscale",
            variable=self.headings["grayscale_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(variable_name="grayscale_cb_var"),
        )
        self.headings["frame_cnt_cb_var"] = BooleanVar()
        self.headings["frame_cnt_cbox"] = Checkbutton(
            self.videos_frm,
            text="Print Frame Count",
            variable=self.headings["frame_cnt_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(variable_name="frame_cnt_cb_var"),
        )
        self.headings["apply_clahe_cb_var"] = BooleanVar()
        self.headings["apply_clahe_cbox"] = Checkbutton(
            self.videos_frm,
            text="Apply CLAHE",
            variable=self.headings["apply_clahe_cb_var"],
            command=lambda: self.inverse_all_cb_ticks(
                variable_name="apply_clahe_cb_var"
            ),
        )

        self.videos_frm.grid(row=1, column=0, sticky=W, padx=5, pady=15)
        self.headings["video_name_col_head"].grid(row=0, column=1, sticky=W, padx=5)
        self.headings["crop_video_col_head"].grid(row=0, column=2, sticky=W, padx=5)
        self.headings["start_time_col_head"].grid(row=0, column=3, sticky=W, padx=5)
        self.headings["end_time_col_head"].grid(row=0, column=4, sticky=W, padx=5)
        self.headings["shorten_all_videos_cbox"].grid(row=0, column=5, sticky=W, padx=5)
        self.headings["video_width_col_head"].grid(row=0, column=6, sticky=W, padx=5)
        self.headings["video_height_col_head"].grid(row=0, column=7, sticky=W, padx=5)
        self.headings["downsample_all_videos_cbox"].grid(
            row=0, column=8, sticky=W, padx=5
        )
        self.headings["fps_col_head"].grid(row=0, column=9, sticky=W, padx=5)
        self.headings["change_fps_all_videos_cbox"].grid(
            row=0, column=10, sticky=W, padx=5
        )
        self.headings["grayscale_cbox"].grid(row=0, column=11, sticky=W, padx=5)
        self.headings["frame_cnt_cbox"].grid(row=0, column=12, sticky=W, padx=5)
        self.headings["apply_clahe_cbox"].grid(row=0, column=13, sticky=W, padx=5)
        self.headings["video_quality_head"].grid(row=0, column=14, sticky=NW)

    def create_video_rows(self):
        self.videos = {}
        for w in self.videos_frm.grid_slaves():
            if w.grid_info()["column"] == 14 and w.grid_info()["row"] > 0:
                w.grid_remove()
        for video_cnt, (name, data) in enumerate(self.videos_in_dir_dict.items()):
            self.videos[name] = {}
            row = video_cnt + 1
            self.videos[name]["video_name_lbl"] = Label(
                self.videos_frm, text=name, width=self.max_char_vid_name + 5
            )
            self.videos[name]["crop_btn"] = Button(
                self.videos_frm,
                text="Crop",
                fg="black",
                command=lambda k=self.videos[name]["video_name_lbl"][
                    "text"
                ]: self.batch_process_crop_function(k),
            )
            self.videos[name]["start_time_var"] = StringVar()
            self.videos[name]["start_time_var"].set("00:00:00")
            self.videos[name]["start_entry"] = Entry(
                self.videos_frm,
                width=6,
                textvariable=self.videos[name]["start_time_var"],
            )
            self.videos[name]["end_time_var"] = StringVar()
            self.videos[name]["end_time_var"].set(data["video_length"])
            self.videos[name]["end_entry"] = Entry(
                self.videos_frm, width=6, textvariable=self.videos[name]["end_time_var"]
            )
            self.videos[name]["clip_cb_var"] = BooleanVar()
            self.videos[name]["clip_cb"] = Checkbutton(
                self.videos_frm, variable=self.videos[name]["clip_cb_var"], command=None
            )
            self.videos[name]["width_var"] = IntVar()
            self.videos[name]["width_var"].set(data["width"])
            self.videos[name]["width_entry"] = Entry(
                self.videos_frm, width=6, textvariable=self.videos[name]["width_var"]
            )
            self.videos[name]["height_var"] = IntVar()
            self.videos[name]["height_var"].set(data["height"])
            self.videos[name]["height_entry"] = Entry(
                self.videos_frm, width=6, textvariable=self.videos[name]["height_var"]
            )
            self.videos[name]["downsample_cb_var"] = BooleanVar()
            self.videos[name]["downsample_cb"] = Checkbutton(
                self.videos_frm,
                variable=self.videos[name]["downsample_cb_var"],
                command=None,
            )
            self.videos[name]["fps_var"] = DoubleVar()
            self.videos[name]["fps_var"].set(round(data["fps"], 4))
            self.videos[name]["fps_entry"] = Entry(
                self.videos_frm, width=6, textvariable=self.videos[name]["fps_var"]
            )
            self.videos[name]["fps_cb_var"] = BooleanVar()
            self.videos[name]["fps_cb"] = Checkbutton(
                self.videos_frm, variable=self.videos[name]["fps_cb_var"], command=None
            )
            self.videos[name]["grayscale_cb_var"] = BooleanVar()
            self.videos[name]["grayscale_cbox"] = Checkbutton(
                self.videos_frm,
                variable=self.videos[name]["grayscale_cb_var"],
                command=None,
            )
            self.videos[name]["frame_cnt_cb_var"] = BooleanVar()
            self.videos[name]["frame_cnt_cbox"] = Checkbutton(
                self.videos_frm,
                variable=self.videos[name]["frame_cnt_cb_var"],
                command=None,
            )
            self.videos[name]["apply_clahe_cb_var"] = BooleanVar()
            self.videos[name]["apply_clahe_cbox"] = Checkbutton(
                self.videos_frm,
                variable=self.videos[name]["apply_clahe_cb_var"],
                command=None,
            )
            self.videos[name]["video_quality_dropdown"] = DropDownMenu(
                self.videos_frm, "", self.cpu_video_quality, "5"
            )
            self.videos[name]["video_quality_dropdown"].setChoices(100)

            self.videos[name]["video_name_lbl"].grid(
                row=row, column=1, sticky=W, padx=5
            )
            self.videos[name]["crop_btn"].grid(row=row, column=2, padx=5)
            self.videos[name]["start_entry"].grid(row=row, column=3, padx=5)
            self.videos[name]["end_entry"].grid(row=row, column=4, padx=5)
            self.videos[name]["clip_cb"].grid(row=row, column=5, sticky=W, padx=5)
            self.videos[name]["width_entry"].grid(row=row, column=6, padx=5)
            self.videos[name]["height_entry"].grid(row=row, column=7, padx=5)
            self.videos[name]["downsample_cb"].grid(row=row, column=8, sticky=W, padx=5)
            self.videos[name]["fps_entry"].grid(row=row, column=9, padx=5)
            self.videos[name]["fps_cb"].grid(row=row, column=10, sticky=W, padx=5)
            self.videos[name]["grayscale_cbox"].grid(
                row=row, column=11, sticky=W, padx=5
            )
            self.videos[name]["frame_cnt_cbox"].grid(
                row=row, column=12, sticky=W, padx=5
            )
            self.videos[name]["apply_clahe_cbox"].grid(
                row=row, column=13, sticky=W, padx=5
            )
            try:
                self.videos[name]["video_quality_dropdown"].grid_remove(
                    row=row, column=14, sticky=NW
                )
            except:
                pass
            self.videos[name]["video_quality_dropdown"].grid(
                row=row, column=14, sticky=NW
            )

    def create_execute_btn(self):
        self.execute_frm = LabelFrame(
            self.main_frm,
            text="EXECUTE",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.reset_all_btn = Button(
            self.execute_frm,
            text="RESET ALL",
            fg="red",
            command=lambda: self.create_video_rows(),
        )
        self.reset_crop_btn = Button(
            self.execute_frm,
            text="RESET CROP",
            fg="orange",
            command=lambda: self.reset_crop(),
        )
        self.execute_btn = Button(
            self.execute_frm, text="EXECUTE", fg="blue", command=lambda: self.execute()
        )

        self.execute_frm.grid(row=2, column=0, sticky=W, padx=5, pady=30)
        self.reset_all_btn.grid(row=0, column=0, sticky=W, padx=5)
        self.reset_crop_btn.grid(row=0, column=1, sticky=W, padx=5)
        self.execute_btn.grid(row=0, column=2, sticky=W, padx=5)

    def reset_crop(self):
        self.crop_dict = {}
        for video_name, video_data in self.videos_in_dir_dict.items():
            self.videos[video_name]["crop_btn"].configure(fg="black")

    def batch_process_crop_function(self, video_name):
        check_file_exist_and_readable(self.videos_in_dir_dict[video_name]["file_path"])
        roi_selector = ROISelector(
            path=self.videos_in_dir_dict[video_name]["file_path"],
            title=f"CROP {video_name} - Press ESC when ROI drawn",
        )
        roi_selector.run()
        self.crop_dict[video_name] = {}
        self.crop_dict[video_name]["top_left_x"] = roi_selector.top_left[0]
        self.crop_dict[video_name]["top_left_y"] = roi_selector.top_left[1]
        self.crop_dict[video_name]["width"] = roi_selector.width
        self.crop_dict[video_name]["height"] = roi_selector.height
        self.crop_dict[video_name]["bottom_right_x"] = roi_selector.bottom_right[0]
        self.crop_dict[video_name]["bottom_right_y"] = roi_selector.bottom_right[1]
        k = cv2.waitKey(20) & 0xFF
        cv2.destroyAllWindows()
        self.videos[video_name]["crop_btn"].configure(fg="red")

    def execute(self):
        out_video_dict = {}
        out_video_dict["meta_data"] = {}
        out_video_dict["video_data"] = {}
        out_video_dict["meta_data"]["in_dir"] = self.input_dir
        out_video_dict["meta_data"]["out_dir"] = self.output_dir
        out_video_dict["meta_data"]["gpu"] = self.use_gpu_var.get()
        if self.use_gpu_var.get() and not check_nvidea_gpu_available():
            raise FFMPEGCodecGPUError(
                msg="No GPU found (as evaluated by nvidea-smi returning None)",
                source=self.__class__.__name__,
            )
        for video_cnt, (name, data) in enumerate(self.videos_in_dir_dict.items()):
            out_video_dict["video_data"][name] = {}
            out_video_dict["video_data"][name]["video_info"] = self.videos_in_dir_dict[
                name
            ]
            if not self.use_gpu_var.get():
                out_video_dict["video_data"][name]["output_quality"] = (
                    self.percent_to_crf_lookup[
                        self.videos[name]["video_quality_dropdown"].getChoices()
                    ]
                )
            else:
                out_video_dict["video_data"][name]["output_quality"] = (
                    self.video_quality_to_preset_lookup[
                        self.videos[name]["video_quality_dropdown"].getChoices()
                    ]
                )
            if name in self.crop_dict.keys():
                out_video_dict["video_data"][name]["crop"] = True
                out_video_dict["video_data"][name]["crop_settings"] = self.crop_dict[
                    name
                ]
            else:
                out_video_dict["video_data"][name]["crop"] = False
                out_video_dict["video_data"][name]["crop_settings"] = None
            if self.videos[name]["clip_cb_var"].get():
                out_video_dict["video_data"][name]["clip"] = True
                out_video_dict["video_data"][name]["clip_settings"] = {
                    "start": self.videos[name]["start_time_var"].get(),
                    "stop": self.videos[name]["end_time_var"].get(),
                }
            else:
                out_video_dict["video_data"][name]["clip"] = False
                out_video_dict["video_data"][name]["clip_settings"] = None
            if self.videos[name]["downsample_cb_var"].get():
                out_video_dict["video_data"][name]["downsample"] = True
                out_video_dict["video_data"][name]["downsample_settings"] = {
                    "width": self.videos[name]["width_var"].get(),
                    "height": self.videos[name]["height_var"].get(),
                }
            else:
                out_video_dict["video_data"][name]["downsample"] = False
                out_video_dict["video_data"][name]["downsample_settings"] = None
            if self.videos[name]["fps_cb_var"].get():
                out_video_dict["video_data"][name]["fps"] = True
                out_video_dict["video_data"][name]["fps_settings"] = {
                    "fps": self.videos[name]["fps_var"].get()
                }
            else:
                out_video_dict["video_data"][name]["fps"] = False
                out_video_dict["video_data"][name]["fps_settings"] = None
            if self.videos[name]["grayscale_cb_var"].get():
                out_video_dict["video_data"][name]["grayscale"] = True
                out_video_dict["video_data"][name]["grayscale_settings"] = None
            else:
                out_video_dict["video_data"][name]["grayscale"] = False
                out_video_dict["video_data"][name]["grayscale_settings"] = None
            if self.videos[name]["frame_cnt_cb_var"].get():
                out_video_dict["video_data"][name]["frame_cnt"] = True
                out_video_dict["video_data"][name]["frame_cnt_settings"] = None
            else:
                out_video_dict["video_data"][name]["frame_cnt"] = False
                out_video_dict["video_data"][name]["frame_cnt_settings"] = None
            if self.videos[name]["apply_clahe_cb_var"].get():
                out_video_dict["video_data"][name]["clahe"] = True
                out_video_dict["video_data"][name]["clahe_settings"] = None
            else:
                out_video_dict["video_data"][name]["clahe"] = False
                out_video_dict["video_data"][name]["clahe_settings"] = None
            out_video_dict["video_data"][name]["last_operation"] = None
            for operation in [
                "clahe",
                "frame_cnt",
                "grayscale",
                "fps",
                "downsample",
                "clip",
                "crop",
            ]:
                if out_video_dict["video_data"][name][operation]:
                    out_video_dict["video_data"][name]["last_operation"] = operation

        self.save_path = os.path.join(self.output_dir, "batch_process_log.json")
        with open(self.save_path, "w") as fp:
            json.dump(out_video_dict, fp)
        self.perform_unit_tests(out_video_dict["video_data"])

    def perform_unit_tests(self, out_video_dict):
        timer = SimbaTimer(start=True)
        for video_name, video_data in out_video_dict.items():
            if video_data["crop"]:
                check_int(
                    value=video_data["crop_settings"]["width"],
                    min_value=1,
                    name=f"Crop width for video {video_name}",
                )
                check_int(
                    value=video_data["crop_settings"]["height"],
                    min_value=1,
                    name=f"Crop height for video {video_name}",
                )
            if video_data["clip"]:
                for variable in ["start", "stop"]:
                    check_if_string_value_is_valid_video_timestamp(
                        value=video_data["clip_settings"][variable],
                        name=f"Clip {variable} time for video {video_name}",
                    )
                check_that_hhmmss_start_is_before_end(
                    start_time=video_data["clip_settings"]["start"],
                    end_time=video_data["clip_settings"]["stop"],
                    name=f"Clip time for video {video_name}",
                )
                for variable in ["start", "stop"]:
                    check_if_hhmmss_timestamp_is_valid_part_of_video(
                        timestamp=video_data["clip_settings"][variable],
                        video_path=video_data["video_info"]["file_path"],
                    )
            if video_data["downsample"]:
                check_int(
                    value=video_data["downsample_settings"]["width"],
                    min_value=1,
                    name=f"Downsample width for video {video_name}",
                )
                check_int(
                    value=video_data["downsample_settings"]["height"],
                    min_value=1,
                    name=f"Downsample height for video {video_name}",
                )
            if video_data["fps"]:
                check_float(
                    value=video_data["fps_settings"]["fps"],
                    min_value=0,
                    name=f"FPS settings for video {video_name}",
                )

        ffmpeg_runner = FFMPEGCommandCreator(json_path=self.save_path)
        ffmpeg_runner.crop_videos()
        ffmpeg_runner.clip_videos()
        ffmpeg_runner.downsample_videos()
        ffmpeg_runner.apply_fps()
        ffmpeg_runner.apply_grayscale()
        ffmpeg_runner.apply_frame_count()
        ffmpeg_runner.apply_clahe()
        ffmpeg_runner.move_all_processed_files_to_output_folder()
        timer.stop_timer()
        stdout_success(
            msg=f"SimBA batch pre-process JSON saved at {self.save_path}",
            source=self.__class__.__name__,
        )
        stdout_success(
            msg=f"Video batch pre-processing complete, new videos stored in {self.output_dir}",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = BatchProcessFrame(input_dir=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos',
#                          output_dir=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/edited')
# test.create_main_window()
# test.create_video_table_headings()
# test.create_video_rows()
# test.create_execute_btn()
# test.main_frm.mainloop()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        FileSelect)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.printing import stdout_success


class SetMachineModelParameters(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="SET MODEL PARAMETERS")
        ConfigReader.__init__(self, config_path=config_path)
        self.clf_table_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.SET_RUN_ML_PARAMETERS.value,
        )
        Label(
            self.clf_table_frm,
            text="CLASSIFIER",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        ).grid(row=0, column=0)
        Label(
            self.clf_table_frm,
            text="MODEL PATH (.SAV)",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        ).grid(row=0, column=1, sticky=NW)
        Label(
            self.clf_table_frm,
            text="DISCRIMINATION THRESHOLD",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        ).grid(row=0, column=2, sticky=NW)
        Label(
            self.clf_table_frm,
            text="MINIMUM BOUT LENGTH (MS)",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        ).grid(row=0, column=3, sticky=NW)
        self.clf_data = {}
        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.clf_data[clf_name] = {}
            Label(
                self.clf_table_frm,
                text=clf_name,
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
            ).grid(row=clf_cnt + 1, column=0, sticky=NW)
            self.clf_data[clf_name]["path"] = FileSelect(
                self.clf_table_frm, title="Select model (.sav) file"
            )
            self.clf_data[clf_name]["threshold"] = Entry_Box(
                self.clf_table_frm, "", "0"
            )
            self.clf_data[clf_name]["min_bout"] = Entry_Box(self.clf_table_frm, "", "0")
            self.clf_data[clf_name]["path"].grid(row=clf_cnt + 1, column=1, sticky=NW)
            self.clf_data[clf_name]["threshold"].grid(
                row=clf_cnt + 1, column=2, sticky=NW
            )
            self.clf_data[clf_name]["min_bout"].grid(
                row=clf_cnt + 1, column=3, sticky=NW
            )

        set_btn = Button(
            self.main_frm,
            text="SET MODEL(S)",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="red",
            command=lambda: self.set(),
        )
        self.clf_table_frm.grid(row=0, sticky=W, pady=5, padx=5)
        set_btn.grid(row=1, pady=10)

    def set(self):
        for model_name, model_settings in self.clf_data.items():
            check_file_exist_and_readable(model_settings["path"].file_path)
            check_float(
                name="Classifier {} threshhold".format(model_name),
                value=model_settings["threshold"].entry_get,
                max_value=1.0,
                min_value=0.0,
            )
            check_int(
                name="Classifier {} minimum bout".format(model_name),
                value=model_settings["min_bout"].entry_get,
                min_value=0.0,
            )

        for cnt, (model_name, model_settings) in enumerate(self.clf_data.items()):
            self.config.set(
                "SML settings",
                "model_path_{}".format(str(cnt + 1)),
                model_settings["path"].file_path,
            )
            self.config.set(
                "threshold_settings",
                "threshold_{}".format(str(cnt + 1)),
                model_settings["threshold"].entry_get,
            )
            self.config.set(
                "Minimum_bout_lengths",
                "min_bout_{}".format(str(cnt + 1)),
                model_settings["min_bout"].entry_get,
            )

        with open(self.config_path, "w") as f:
            self.config.write(f)

        stdout_success(
            msg="Model paths/settings saved in project_config.ini",
            source=self.__class__.__name__,
        )


# _ = SetMachineModelParameters(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')
